# Library Importation
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import plotly
import plotly_express as px 
from sklearn.linear_model import LinearRegression
from sklearn.decomposition import TruncatedSVD
from sklearn.metrics.pairwise import manhattan_distances
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.utils._mask import _get_mask
from sklearn.metrics.pairwise import check_pairwise_arrays
from sklearn.metrics.pairwise import is_scalar_nan
import time
import warnings
import logging
logging.getLogger('matplotlib').setLevel(logging.ERROR)
warnings.simplefilter("ignore", category=RuntimeWarning)
from sklearn.impute import SimpleImputer


def demean(S, w):
    w_sum = np.matmul(w, S)
    w_rowsum = w_sum.reshape((1, S.shape[0]))
    w_colsum = w_sum.reshape((S.shape[0], 1))
    w_totalsum = np.dot(w_sum, w)
    S -= (w_rowsum + w_colsum) -  w_totalsum
    return S

def weighted_cmdscale(D, weights, num_dims):
    n = len(D)
    H = np.eye(n) - np.ones((n, n))/n
    S = -H.dot(D**2).dot(H)/2
    weights_normalized = weights / weights.sum()
    S = demean(S, weights_normalized)
    W = np.diag(weights)
    WSW = np.matmul(np.matmul(np.sqrt(W), S), np.sqrt(W))
    svd = TruncatedSVD(num_dims, algorithm="arpack")
    svd.fit(WSW)
    dist_transformed = np.matmul(svd.components_, np.linalg.inv(np.sqrt(W))).T
    return dist_transformed

def nan_manhattan_distances(X, Y=None):    
    missing_values=np.nan
    force_all_finite = 'allow-nan' if is_scalar_nan(missing_values) else True
    X, Y = check_pairwise_arrays(X, Y, accept_sparse=False,
                                 force_all_finite=force_all_finite)
    missing_X = _get_mask(X, missing_values)
    missing_Y = missing_X if Y is X else _get_mask(Y, missing_values)
    
    if X is Y:
        X_is_Y = True
    else:
        X_is_Y = False

    X[missing_X] = 0
    Y[missing_Y] = 0

    distances = manhattan_distances(X, Y)

    X = np.abs(X)
    Y = np.abs(Y)
    Y = Y.T
    distances -= np.dot(X, missing_Y.T)
    distances -= np.dot(missing_X, Y)

    np.clip(distances, 0, None, out=distances)

    if X_is_Y:
        np.fill_diagonal(distances, 0.0)

    present_X = ~missing_X
    present_Y = present_X if X_is_Y else ~missing_Y
    X = present_X.astype(np.float32)
    Y = present_Y.T.astype(np.float32)
    present_count = np.dot(X, Y)
    distances[present_count == 0] = np.nan
    np.maximum(1, present_count, out=present_count)
    distances /= present_count

    return distances

def nan_rms_distances(X, Y=None):    
    missing_values=np.nan
    force_all_finite = 'allow-nan' if is_scalar_nan(missing_values) else True
    X, Y = check_pairwise_arrays(X, Y, accept_sparse=False,
                                 force_all_finite=force_all_finite)
    missing_X = _get_mask(X, missing_values)
    missing_Y = missing_X if Y is X else _get_mask(Y, missing_values)

    if X is Y:
        X_is_Y = True
    else:
        X_is_Y = False
    
    X[missing_X] = 0
    Y[missing_Y] = 0

    distances = euclidean_distances(X, Y, squared=True)

    X = X * X
    Y = Y * Y
    Y = Y.T
    distances -= np.dot(X, missing_Y.T)
    distances -= np.dot(missing_X, Y)

    np.clip(distances, 0, None, out=distances)

    if X_is_Y:
        np.fill_diagonal(distances, 0.0)

    present_X = ~missing_X
    present_Y = present_X if X_is_Y else ~missing_Y
    X = present_X.astype(np.float32)
    Y = present_Y.T.astype(np.float32)
    present_count = np.dot(X, Y)
    distances[present_count == 0] = np.nan
    np.maximum(1, present_count, out=present_count)
    distances /= present_count
    
    np.sqrt(distances, out=distances)

    return distances

def nan_ap_distances(X, Y=None):
    missing_values=np.nan
    force_all_finite = 'allow-nan' if is_scalar_nan(missing_values) else True
    X, Y = check_pairwise_arrays(X, Y, accept_sparse=False,
                                 force_all_finite=force_all_finite)
    missing_X = _get_mask(X, missing_values)
    missing_Y = missing_X if Y is X else _get_mask(Y, missing_values)
    
    if X is Y:
        X_is_Y = True
    else:
        X_is_Y = False
    
    X[missing_X] = 0
    Y[missing_Y] = 0

    present_X = ~missing_X
    present_Y = present_X if X_is_Y else ~missing_Y
    
    singlerow_X = np.dot(X, present_Y.T)
    singlerow_Y = np.dot(Y, present_X.T)
    Y = Y.T
    distances = singlerow_X - 2 * np.dot(X, Y) + singlerow_Y.T
    
    if X_is_Y:
        np.fill_diagonal(distances, 0.0)
    
    X = present_X.astype(np.float32)
    Y = present_Y.T.astype(np.float32)
    present_count = np.dot(X, Y)    
    distances[present_count == 0] = np.nan
    np.maximum(1, present_count, out=present_count)
    distances /= present_count
    
    return distances

def distance_mat(first, second=None, dist_func = 'AP'):
    """                                                                                       
    Distance Matrix Computation Function between two different arrays.
    
    Parameters                                                                                
    ----------                                                                                
    first       : (m, n1) array
                  First genome matrix for a chosen ancestry. Indicates which genetic encoding
                  is present for individual n at position m, or NaN if position is masked.
                  
    second      : (m, n2) array
                  Second genome matrix for a chosen ancestry. Indicates which genetic encoding
                  is present for individual n at position m, or NaN if position is masked.
                  
    dist_func   : string
                  Distance function utilized in computation. Defaulted to be 'AP'.
                  Options to choose from are: 'Manhattan', 'RMS', 'AP'
                                                                                
    Returns                                                                                   
    -------                                                                                   
    distance   : (n1, n2) array                                                                       
                 Computed distance between two individuals i of first and j of second provided at entry (i, j).
    """
    start = time.time()
    first = first.T
    first = first.astype(np.float32)
    
    if second is not None:
        second = second.T
        second = second.astype(np.float32)
        if second.shape[0] == 0:
            logging.info("Distance Matrix building: --- %s seconds ---" % (time.time() - start))
            return np.array([[]] * first.shape[0])
        if first.shape[0] == 0:
            logging.info("Distance Matrix building: --- %s seconds ---" % (time.time() - start))
            return np.array([[]] * second.shape[0]).T
        
    if first.shape[0] == 0:
        logging.info("Distance Matrix building: --- %s seconds ---" % (time.time() - start))
        return np.array([[]]).T.dot(np.array([[]]))
    
    if (dist_func == 'Manhattan'):
        distance = nan_manhattan_distances(first, second)   
        
    elif (dist_func == 'RMS'):
        distance = nan_rms_distances(first, second)
        
    elif (dist_func == 'AP'):
        distance = nan_ap_distances(first, second)
            
    else:
        logging.info("Distance Function Undefined -- Defaulted to AP Distance")
        distance = nan_ap_distances(first, second)
    logging.info("Distance Matrix building: --- %s seconds ---" % (time.time() - start))
    
    return distance


def intersection(lst1, lst2): 
    return list(set(lst1) & set(lst2))  

def binary_intersection(rs_ID_list):
    """                                                                                       
    Intersecting Positions Calculation between two arrays:
    
    Parameters                                                                                
    ----------                                                                                
    rs_ID_list  : (num_arrays, ) list 
        List of rs IDs for each of the processed arrays.
                                                                                
    Returns                                                                                   
    -------                                                                                   
    binary   : (num_arrays, num_arrays) list                                                                      
        Entry (i, j) contains an array of intersecting rs IDs between array i,
        and array j.
    """
    binary = []
    for i in range(0,len(rs_ID_list)):
        binary.append([1]*(i+1))
        first = list(rs_ID_list[i])
        for j in range(i+1, len(rs_ID_list)):
            second = list(rs_ID_list[j])
            intersect = intersection(first, second)
            binary[i].append(intersect)
            
    for i in range(len(binary)):
        logging.info("Array %s", i+1)
        for j in range(i+1,len(binary[i])):
            logging.info("Intersect with %s : %s", j+1, len(binary[i][j]))
            
    return binary

def build_overlap(overlap, array1_ID, array2_ID, array1, array2):
    '''
    Build Overlap Function.
    
    Parameters                                                                                
    ----------                                                                                
    overlap  : List of rs IDs in common between the two arrays.
    array1_ID: List of all rs IDs in the first array.
    array2_ID: List of all rs IDs in the second array.
    array1   : Masked genetic matrix for the first array.
    array2   : Masked genetic matrix for the second array.

    Returns                                                                                   
    -------                                                                                   
    overlap_mat   : array representing the combined genetic matrix of 
                    individuals in both arrays across their common snps.
    '''
    index_1 = np.where(np.in1d(array1_ID, overlap))[0]
    index_2 = np.where(np.in1d(array2_ID, overlap))[0]

    return [array1[index_1],array2[index_2]]


def overlap_blocks(ancestry, ref_col, ref_row, num_arrays, rs_ID_list, binary, masks):
    """                                                                                       
    Binary Intersection Array Computation:
    
    Parameters                                                                                
    ---------- 
    ancestry    : integer
                  Ancestry of interest to perform distance matrix computation on.
    ref_col     : integer
                  Column of reference for conversion calculation.
    ref_row     : integer
                  Row of reference for conversion calculation.
    num_arrays  : integer
                  Number of arrays in dataset.
    rs_ID_list  : (num_arrays, ) list 
                  List of rs IDs for each of the processed arrays.
    binary      : (num_arrays, num_arrays) list                                                                      
                  Entry (i, j) contains an array of intersecting rs IDs between array i,
                  and array j.
    masks       : (num_arrays, ) list                                                                          
                  List of masked matrices for each ancestries at each given array.
                                                                   
    Returns                                                                                   
    -------                                                                                   
    overlap   : (num_arrays, num_arrays) list
                Sets of overlapping matrices between two arrays across common snps
    """    
    overlap = []
    for i in range(num_arrays):
        overlap.append([1]*i)
        for j in range(i, num_arrays):
            if (j == 0):
                ind = np.where(np.in1d(rs_ID_list[j], np.array(binary[ref_row][ref_col])))[0]
                overlap[i].append(masks[j][ancestry][ind, :])
            elif (j == i):
                ind = np.where(np.in1d(rs_ID_list[j], binary[ref_row][j]))[0]
                overlap[i].append(masks[j][ancestry][ind, :])
            else:
                overlap[i].append(build_overlap(np.array(binary[i][j]), rs_ID_list[i], rs_ID_list[j], masks[i][ancestry], masks[j][ancestry]))
    return overlap


def convert(overlap_1, overlap_2, array, rs_IDs, distance_type, plot_reg, index):
    '''
    Build Overlap Function.
    
    Parameters                                                                                
    ----------                                                                                
    overlap_1      : First set of overlapping positions between two arrays.
    overlap_2      : Second set of overlapping positions between two arrays.
    array          : Combined matrix of two arrays individuals across common snps.
    rs_IDs         : List of all positions present in array. 
    distance_type  : Distance function utilized in computation. Defaulted to be 'AP'.
                     Options to choose from are: 'Manhattan', 'RMS', 'AP'

    Returns                                                                                   
    -------                                                                                   
    [reg.coef_, reg.intercept_]   : conversion matrix for distance matrix.
    
    '''
    ind1 = np.where(np.in1d(rs_IDs, overlap_1))[0]
    ind2 = np.where(np.in1d(rs_IDs, overlap_2))[0]   
    
    logging.info("--- Number of overlapping positions: --- %s", len(ind1))
    dist1 = distance_mat(first=array[ind1], dist_func=distance_type)
    dist2 = distance_mat(first=array[ind2], dist_func=distance_type)

    dist1 = np.reshape(dist1, (dist1.size, 1)).flatten()
    dist2 = np.reshape(dist2, (dist2.size, 1)).flatten()
    
    remove1 = np.argwhere(np.isnan(dist1)).flatten()
    remove2 = np.argwhere(np.isnan(dist2)).flatten()
    remove = np.array(list(set(remove1) | set(remove2)), dtype = np.int32)
    
    dist1 = np.delete(dist1, remove)
    dist2 = np.delete(dist2, remove)
    
    dist1 = np.reshape(dist1, (dist1.size, 1))
    dist2 = np.reshape(dist2, (dist2.size, 1))

    reg = LinearRegression().fit(dist1, dist2)
    
    if (plot_reg):
        plt.scatter(dist1, dist2, marker='x')
        plt.plot(dist1, dist1 * reg.coef_ + reg.intercept_)
        plt.savefig('conv_plot_' + index + '.jpg')
        plt.clf()

    return [reg.coef_, reg.intercept_, reg.score(dist1,dist2)]

def conversion_metrics(ancestry, ref_col, ref_row, num_arrays, rs_ID_list, binary, masks, distance_type, plot_reg=False):
    """                                                                                       
    Conversion Matrix Computation Function:
    
    Parameters                                                                                
    ---------- 
    ancestry       : integer
                    Ancetry of interest to perform distance matrix computation on.
    ref_col        : integer
                    Column of reference for conversion calculation.
    ref_row        : integer
                     Row of reference for conversion calculation.
    num_arrays     : integer
                     Number of arrays in dataset.
    rs_ID_list     : (num_arrays, ) list 
                     List of rs IDs for each of the processed arrays.
    binary         : (num_arrays, num_arrays) list                                                                      
                    Entry (i, j) contains an array of intersecting rs IDs between array i,
                    and array j.
    masks          : (num_arrays, ) list                                                                          
                     List of masked matrices for each ancestries at each given array.
    distance_type  : string
                     Distance function utilized in computation. Defaulted to be 'AP'.
                     Options to choose from are: 'Manhattan', 'RMS', 'AP'
                                                                   
    Returns                                                                                   
    -------                                                                                   
    conversion  : (num_arrays, num_arrays) list
                  Conversion coefficients for each of the arrays in our dataset.
    intercept   : (num_arrays, num_arrays) list
                  Conversion intercepts for each of the arrays in our dataset.
    """
    conversion = np.ones((num_arrays,num_arrays))
    intercept = np.zeros((num_arrays,num_arrays))
    score = np.zeros((num_arrays,num_arrays))
    for j in range(ref_col+1, num_arrays):
        for i in range(j+1):
            index = str(10*j+i)
            if (i == ref_col):
                conv_list = convert(binary[i][j], binary[ref_row][ref_col], masks[i][ancestry], rs_ID_list[i], distance_type, plot_reg, index)
                conversion[i][j] = conv_list[0]
                intercept[i][j] = conv_list[1]
                score[i][j] = conv_list[2]
            elif (i == ref_row):
                conv_list = convert(binary[i][j], binary[ref_row][ref_col], masks[i][ancestry], rs_ID_list[i], distance_type, plot_reg, index)
                conversion[i][j] = conv_list[0]
                intercept[i][j] = conv_list[1]                
                score[i][j] = conv_list[2]
            elif (i != j):
                conv_list = convert(binary[i][j], binary[ref_row][j], masks[j][ancestry], rs_ID_list[j], distance_type, plot_reg, index)
                conversion[i][j] = conv_list[0]
                intercept[i][j] = conv_list[1]
                score[i][j] = conv_list[2]
            else:
                conversion[i][j] = conversion[ref_row][j]
                intercept[i][j] = intercept[ref_row][j]
                score[i][j] = conv_list[2]

    # Symmetric:
    for i in range(conversion.shape[0]):
        for j in range(conversion.shape[1]):
            conversion[j][i] = conversion[i][j]
            intercept[j][i] = intercept[i][j]
            score[j][i] = score[i][j]
    
    np.savetxt('conversion.csv', conversion, delimiter=',')
    np.savetxt('intercept.csv', intercept, delimiter=',')
    np.savetxt('score.csv', score, delimiter=',')

    return conversion, intercept

def distance_overlap(ref_col, ref_row, num_arrays, overlap, conversion, intercept, distance_type):
    """                                                                                       
    Binary Intersection Distance Matrix Computation:
    
    Parameters                                                                                
    ---------- 
    ref_col     : integer
                  Column of reference for conversion calculation.
    ref_row     : integer
                  Row of reference for conversion calculation.
    num_arrays  : integer
                  Number of arrays in dataset.
    overlap     : (num_arrays, num_arrays) list
                  Sets of overlapping matrices between two arrays across common snps
    conversion  : (num_arrays, num_arrays) list
                  Conversion coefficient estimated from linear fit, to account for the 
                  difference in chosen snps between any two arrays, and the reference
                  arrays determined with a row and column.
    intercept   : (num_arrays, num_arrays) list
                  Conversion intercept estimated from linear fit, to account for the 
                  difference in chosen snps between any two arrays, and the reference
                  arrays determined with a row and column.
    distance_type : string
                  Distance function utilized in computation. Defaulted to be 'AP'.
                  Options to choose from are: 'Manhattan', 'RMS', 'AP'
                                                                                
    Returns                                                                                   
    -------                                                                                   
    dist_lis   : (num_arrays, num_arrays) list                                                                      
        Entry (i, j) contains a distance matrix array of intersecting rs IDs between array i,
        and array j.
    """
    distance_list = []
    for i in range(len(overlap)):
        distance_row = []
        for j in range(len(overlap[0])):
            if (j > i):
                distance_row.append(distance_mat(first=overlap[i][j][0], second=overlap[i][j][1], dist_func=distance_type))
            elif (i == j):
                distance_row.append(distance_mat(first=overlap[i][j], dist_func=distance_type))
            else:
                distance_row.append(1)
        distance_list.append(distance_row)

    # Make distance_list symmetric
    for i in range(len(distance_list)):
        for j in range(len(distance_list[0])):
            if (j < i):
                distance_list[i][j] = distance_list[j][i]
                
#     dist_lis = [[distance_list[0][0],distance_list[0][1]],[distance_list[1][0],distance_list[1][1]], [1, 1], [1, 1, 1]]
    dist_lis = [[distance_list[0][0],distance_list[0][1]],[distance_list[1][0],distance_list[1][1]]]
    for i in range(2,num_arrays):
        dist_lis.append(i*[1])
        
    for j in range(ref_col+1, num_arrays):
        for i in range(j+1):
            if (i == ref_col) or (i == ref_row):
                dist_lis[i].append(conversion[i][j]*distance_list[i][j] + intercept[i][j])
            elif (i != j):
                dist_lis[i].append((distance_list[i][j]*conversion[i][j] + intercept[i][j])*conversion[ref_row][j]+intercept[ref_row][j])
            else:
                dist_lis[i].append(conversion[ref_row][j]*distance_list[i][j] + intercept[ref_row][j])

    # Make distance_list symmetric
    for i in range(len(dist_lis)):
        for j in range(len(dist_lis[0])):
            if (j < i):
                dist_lis[i][j] = dist_lis[j][i].T  
                
    return dist_lis

def combine_dist_mat(distance_list):
    """                                                                                       
    Combine Distance Matrix Function:
    
    Parameters                                                                                
    ---------- 
    distance_list: (num_arrays, num_arrays) list 
                   List of computed distances on individuals of two intersecting arrays 
                   over the same set of snp's.
                                                                                
    Returns                                                                                   
    -------                                                                                   
    dist_mat     : (m, m)                                                                     
                   Combined and converted distance matrix of all individuals in the dataset.
    """
    # Construct Full Distance Matrix over selected individuals of overlap
    dist_mat = distance_list[0][0]
    for i in range(1,len(distance_list[0])):
        dist_mat = np.hstack((dist_mat, distance_list[0][i]))#*conversion[0][i]))

    for i in range(1,len(distance_list)):
        dist_row = distance_list[i][0]
        for j in range(1,len(distance_list[i])):
            if (i == j): 
                dist_row = np.hstack((dist_row, distance_list[i][j]))
            elif (i < j):
                dist_row = np.hstack((dist_row, distance_list[i][j]))
            else: 
                dist_row = np.hstack((dist_row, distance_list[i][j]))
        dist_mat = np.vstack((dist_mat, dist_row))
    return dist_mat

def mean_impute(d):
    imp = SimpleImputer(missing_values=np.nan, strategy='mean')
    imp.fit(d)
    d = imp.transform(d)
    return d

def additive_impute(d):
    num_missing = np.isnan(d).sum()
    n = d.shape[0]
    while True:
        old_num_missing = num_missing
        for i in range(n):
            for j in range(n):
                if np.isnan(d[i,j]):
                    MinMax = np.nanmax(d)
                    fill_d = False
                    for k in range(n):
                        for l in range(n):
                            if (np.isnan([d[i,k], d[j,k], d[i,l], d[j,l], d[k,l]]) == False).all():
                                fill_d = True
                                Max = max((d[i,k] + d[j,l]), (d[i,l] + d[j,k])) - d[k,l]
                                if Max < MinMax:
                                    MinMax = Max
                    if fill_d:
                        d[i,j] = MinMax
                        num_missing -= 1
        if num_missing == old_num_missing:
            break
    return d

def remove_individuals(dist_mat, groups, weights, ind_IDs):
    keep_inds = np.argwhere(np.isnan(dist_mat).sum(axis=0) != dist_mat.shape[0]).flatten()
    dist_mat = dist_mat[keep_inds, :][:, keep_inds]
    groups = groups[keep_inds]
    weights = weights[keep_inds]
    ind_IDs = ind_IDs[keep_inds]
    return dist_mat, groups, weights, ind_IDs

def mds_transform(distance_list, groups, weights, ind_ID_list, num_dims, num_arrays=1):
    """                                                                                       
    Multi-Dimensional Scaling (MDS) Function:
    
    Parameters                                                                                
    ---------- 
    distance_list: (num_arrays, num_arrays) list 
                   List of computed distances on all individuals in the dataset. 
    groups       : (n,) array
                   Corresponding groups for each individual in the dataset.
                   Order is consistent with the order of individuals in the distance matrix.
    weights      : (n,) array
                   Weight assigned to each individual when performing weighted classical MDS.
    ind_ID_list  : list of str
                   List of individual IDs for each of the processed arrays.
    num_dims     : int
                   Number of MDS dimensions to return.
    num_arrays   : int
                   Total number of arrays in the dataset.
                                                                                 
    Returns                                                                                   
    -------                                                                                   
    np.ndarray of shape (n, num_dims)
        The MDS embedding for the n individuals across the requested dimensions.
    """
    ind_IDs = ind_ID_list
    for i in range(1,num_arrays):
        ind_IDs = np.hstack((ind_IDs, ind_ID_list[i]))
    
    dist_mat = combine_dist_mat(distance_list)
    dist_mat, groups, weights, ind_IDs = remove_individuals(dist_mat, groups, weights, ind_IDs)
    np.savetxt('distance_unimputed.csv', dist_mat, delimiter=',')
    
    dist_mat = additive_impute(dist_mat)
    np.savetxt('distance_imputed.csv', dist_mat, delimiter=',')
    dist_transformed = weighted_cmdscale(dist_mat, weights, num_dims)
    
    categories = pd.factorize(pd.Series(groups))
    
    output_df = pd.DataFrame()
    output_df['indID'] = ind_IDs
    output_df['label'] = groups
    for i in range(num_dims):
        output_df['MDS' + str(i+1)] = dist_transformed[:,i]
    output_df.to_csv('output.tsv', sep='\t', index=False)

    return dist_transformed[:, :num_dims]


import subprocess
from typing import List, Optional, Dict, Any
import os
import sys
import argparse
import shutil


class PongVizError(Exception):
    """Custom exception for Pong visualization errors"""
    pass


class FileMapError(Exception):
    """Custom exception for filemap creation errors"""
    pass


def pong_viz(folder_runs: str, output_dir: str, k: Optional[int] = None, min_k: Optional[int] = None, max_k: Optional[int] = None,
             runs: List[int] = None, run_prefix: str = 'train', ind2pop_path: Optional[str] = None,
             pop_names_path: Optional[str] = None, color_list_path: Optional[str] = None, verbose: bool = False) -> None:
    """
    Executes Pong visualization with the specified parameters.
    """

    # Check if pong is installed
    if not shutil.which('pong'):  # TODO: replace pong dependency with a custom implementation
        raise PongVizError("Pong is not installed. Please install it using 'pip install pong'")

    # Validate input parameters
    if not os.path.isdir(folder_runs):
        raise PongVizError(f"Input folder does not exist: {folder_runs}")
    if not output_dir:
        raise PongVizError("Output directory must be specified")
    os.makedirs(output_dir, exist_ok=True)

    # Optional files validation
    optional_files = {'ind2pop_path': ind2pop_path, 'pop_names_path': pop_names_path, 'color_list_path': color_list_path}
    for name, path in optional_files.items():
        if path and not os.path.isfile(path):
            raise PongVizError(f"Specified {name} file does not exist: {path}")

    try:
        filemap_path = create_filemap(folder_runs, k, min_k, max_k, runs, run_prefix)

        if verbose:
            print(f"Created filemap at: {filemap_path}")

        # Build command
        cmd = ["pong", "-m", filemap_path, "-o", output_dir]
        if ind2pop_path:
            cmd.extend(["-i", ind2pop_path])
        if pop_names_path:
            cmd.extend(["-n", pop_names_path])
        if color_list_path:
            cmd.extend(["-l", color_list_path])

        if verbose:
            print("Executing command:", " ".join(cmd))

        # Execute pong
        subprocess.run(cmd)

        if verbose:
            print("Pong execution successful")
            print("Results saved in:", output_dir)

    except FileMapError as e:
        raise PongVizError(f"Error creating filemap: {str(e)}")
    except subprocess.CalledProcessError as e:
        raise PongVizError(f"Error executing Pong: {str(e)}\nOutput: {e.output}")
    except Exception as e:
        raise PongVizError(f"Unexpected error: {str(e)}")


def create_filemap(folder: str, k: Optional[int] = None, min_k: Optional[int] = None, max_k: Optional[int] = None,
                   runs: List[int] = None, run_prefix: str = 'train_demo',) -> str:
    """
    Creates a filemap for training files organized by k values and runs and saves it to a file.

    Args:
        folder (str): Base folder path
        k (Optional[int]): Single k value to process. If specified, min_k and max_k are ignored
        min_k (Optional[int]): Minimum k value for range processing
        max_k (Optional[int]): Maximum k value for range processing
        runs (List[int]): List of run numbers
        run_prefix (str): Prefix for the run files (default: 'train_demo')

    Returns:
        str: Path to saved file

    Raises:
        FileMapError: If invalid parameters are provided or if configuration is incorrect
    """
    # Input validation
    if not folder:
        raise FileMapError("Folder path must be specified")
    if not runs:
        raise FileMapError("Runs list must be provided and non-empty")
    if not all(isinstance(run, int) and run > 0 for run in runs):
        raise FileMapError("All runs must be positive integers")
    if k is not None:
        if k < 1:
            raise FileMapError("k must be a positive integer")
        k_values = [k]
    else:
        if min_k is None or max_k is None:
            raise FileMapError("Either k or both min_k and max_k must be specified")
        if min_k > max_k:
            raise FileMapError("min_k cannot be greater than max_k")
        if min_k < 1:
            raise FileMapError("min_k must be a positive integer")
        k_values = range(min_k, max_k + 1)

    # Write file:
    filemap = {}
    try:
        for i, k_val in enumerate(k_values):
            for run_num in range(runs[i]):
                key = f'k{k_val}r{run_num+1}'
                if folder != '.':
                    path = folder + '/' + f'run{run_num+1}' + '/' + f'{run_prefix}.{k_val}.Q'
                else:
                    path = f'run{run_num+1}' + '/' + f'{run_prefix}.{k_val}.Q'

                filemap[key] = (k_val, path)

        filemap_path = os.path.join('pong_filemap')

        with open(filemap_path, 'w') as f:
            for i, (key, (k_val, path)) in enumerate(filemap.items()):
                f.write(f"{key}\t{k_val}\t{path}\n")

                if k_val in k_values:
                    index = k_values.index(k_val)
                    if (i + 1) % runs[index] == 0:
                        f.write("\n")

        return filemap_path

    except Exception as e:
        raise FileMapError(f"Error creating filemap: {str(e)}")


def parse_pong_args() -> Dict[str, Any]:
    """
    Parse command line arguments for pong visualization.

    Returns:
        Dict[str, Any]: Dictionary containing all parsed arguments ready for pong_viz

    Raises:
        SystemExit: If argument parsing fails or --help is used
    """
    parser = argparse.ArgumentParser(
        description='Execute Pong visualization with specified parameters',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument('folder_runs', help='Base folder containing the run files')
    parser.add_argument('output_dir', help='Directory where results will be saved')

    k_group = parser.add_mutually_exclusive_group(required=True)
    k_group.add_argument('-k', '--k', type=int, help='Single k value to process')
    k_group.add_argument('--k-range', nargs=2, type=int, metavar=('MIN_K', 'MAX_K'), help='Range of k values (min_k max_k)')

    parser.add_argument('--runs', type=int, nargs='+', help='Number of runs for each K')
    parser.add_argument('--prefix', default='train_demo', help='Prefix for run files')
    parser.add_argument('-i', '--ind2pop', help='Path to individual-to-population mapping file')
    parser.add_argument('-n', '--pop-names', help='Path to population names file')
    parser.add_argument('-c', '--color-list', help='Path to color list file')
    parser.add_argument('-v', '--verbose', action='store_true', help='Print detailed information during execution')
    args = parser.parse_args()

    params = {'folder_runs': args.folder_runs,
              'output_dir': args.output_dir,
              'runs': args.runs,
              'run_prefix': args.prefix,
              'verbose': args.verbose}

    if args.k is not None:
        params['k'] = args.k
    else:
        params['min_k'] = args.k_range[0]
        params['max_k'] = args.k_range[1]

    if args.ind2pop:
        params['ind2pop_path'] = args.ind2pop
    if args.pop_names:
        params['pop_names_path'] = args.pop_names
    if args.color_list:
        params['color_list_path'] = args.color_list

    return params


if __name__ == "__main__":
    try:
        params = parse_pong_args()
        pong_viz(**params)

    except (FileMapError, PongVizError) as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"Unexpected error: {e}", file=sys.stderr)
        sys.exit(1)



import logging
from pathlib import Path
import numpy as np
import copy
import warnings
from typing import Union, List, Dict, Sequence, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from snputils.snp.genobj.snpobj import SNPObject

from .base import AncestryObject

log = logging.getLogger(__name__)


class LocalAncestryObject(AncestryObject):
    """
    A class for window-level Local Ancestry Inference (LAI) data.
    """
    def __init__(
        self,
        haplotypes: List[str], 
        lai: np.ndarray,
        samples: Optional[List[str]] = None, 
        ancestry_map: Optional[Dict[str, str]] = None, 
        window_sizes: Optional[np.ndarray] = None,
        centimorgan_pos: Optional[np.ndarray] = None,
        chromosomes: Optional[np.ndarray] = None,
        physical_pos: Optional[np.ndarray] = None
    ) -> None:
        """
        Args:
            haplotypes (list of str of length n_haplotypes):
                A list of unique haplotype identifiers.
            lai (array of shape (n_windows, n_haplotypes)): 
                A 2D array containing local ancestry inference values, where each row represents a 
                genomic window, and each column corresponds to a haplotype phase for each sample.
            samples (list of str of length n_samples, optional):
                A list of unique sample identifiers.
            ancestry_map (dict of str to str, optional):
                A dictionary mapping ancestry codes to region names.
            window_sizes (array of shape (n_windows,), optional): 
                An array specifying the number of SNPs in each genomic window.
            centimorgan_pos (array of shape (n_windows, 2), optional): 
                A 2D array containing the start and end centimorgan positions for each window.
            chromosomes (array of shape (n_windows,), optional): 
                An array with chromosome numbers corresponding to each genomic window.
            physical_pos (array of shape (n_windows, 2), optional): 
                A 2D array containing the start and end physical positions for each window.
        """
        if lai.ndim != 2:
            raise ValueError("`lai` must be a 2D array with shape (n_windows, n_haplotypes).")
        
        # Determine the number of unique ancestries and samples from the LAI array
        n_ancestries = len(np.unique(lai))
        n_haplotypes = lai.shape[1]
        n_samples = n_haplotypes // 2

        super(LocalAncestryObject, self).__init__(n_samples, n_ancestries)

        self.__haplotypes = haplotypes
        self.__lai = lai
        self.__window_sizes = window_sizes
        self.__centimorgan_pos = centimorgan_pos
        self.__samples = samples
        self.__chromosomes = chromosomes
        self.__physical_pos = physical_pos
        self.__ancestry_map = ancestry_map

        # Perform sanity check to ensure all unique ancestries in LAI data are represented in the ancestry map
        self._sanity_check()

    def __getitem__(self, key):
        """
        To access an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    def __setitem__(self, key, value):
        """
        To set an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    @property
    def haplotypes(self) -> List[str]:
        """
        Retrieve `haplotypes`.

        Returns:
            **list of length n_haplotypes:** A list of unique haplotype identifiers.
        """
        return self.__haplotypes

    @haplotypes.setter
    def haplotypes(self, x):
        """
        Update `haplotypes`.
        """
        self.__haplotypes = x

    @property
    def lai(self) -> np.ndarray:
        """
        Retrieve `lai`.

        Returns:
            **array of shape (n_windows, n_haplotypes):** 
                A 2D array containing local ancestry inference values, where each row represents a 
                genomic window, and each column corresponds to a haplotype phase for each sample.
        """
        return self.__lai

    @lai.setter
    def lai(self, x):
        """
        Update `lai`.
        """
        self.__lai = x

    @property
    def samples(self) -> Optional[List[str]]:
        """
        Retrieve `samples`.

        Returns:
            **list of str:** A list of unique sample identifiers.
        """
        if self.__samples is not None:
            return self.__samples
        elif self.__haplotypes is not None:
            return [hap.split('.')[0] for hap in self.__haplotypes][::2]
        else:
            return None
    
    @samples.setter
    def samples(self, x):
        """
        Update `samples`.
        """
        self.__samples = x

    @property
    def ancestry_map(self) -> Optional[Dict[str, str]]:
        """
        Retrieve `ancestry_map`.

        Returns:
            **dict of str to str:** A dictionary mapping ancestry codes to region names.
        """
        return self.__ancestry_map

    @ancestry_map.setter
    def ancestry_map(self, x):
        """
        Update `ancestry_map`.
        """
        self.__ancestry_map = x

    @property
    def window_sizes(self) -> Optional[np.ndarray]:
        """
        Retrieve `window_sizes`.

        Returns:
            **array of shape (n_windows,):** 
                An array specifying the number of SNPs in each genomic window.
        """
        return self.__window_sizes
        
    @window_sizes.setter
    def window_sizes(self, x):
        """
        Update `window_sizes`.
        """
        self.__window_sizes = x

    @property
    def centimorgan_pos(self) -> Optional[np.ndarray]:
        """
        Retrieve `centimorgan_pos`.

        Returns:
            **array of shape (n_windows, 2):** 
                A 2D array containing the start and end centimorgan positions for each window.
        """
        return self.__centimorgan_pos

    @centimorgan_pos.setter
    def centimorgan_pos(self, x):
        """
        Update `centimorgan_pos`.
        """
        self.__centimorgan_pos = x

    @property
    def chromosomes(self) -> Optional[np.ndarray]:
        """
        Retrieve `chromosomes`.

        Returns:
            **array of shape (n_windows,):** 
                An array with chromosome numbers corresponding to each genomic window.
        """
        return self.__chromosomes
        
    @chromosomes.setter
    def chromosomes(self, x):
        """
        Update `chromosomes`.
        """
        self.__chromosomes = x

    @property
    def physical_pos(self) -> Optional[np.ndarray]:
        """
        Retrieve `physical_pos`.

        Returns:
            **array of shape (n_windows, 2):** 
                A 2D array containing the start and end physical positions for each window.
        """
        return self.__physical_pos

    @physical_pos.setter
    def physical_pos(self, x):
        """
        Update `physical_pos`.
        """
        self.__physical_pos = x

    @property
    def n_samples(self) -> int:
        """
        Retrieve `n_samples`.

        Returns:
            **int:** 
                The total number of samples.
        """
        if self.__samples is not None:
            return len(self.__samples)
        elif self.__haplotypes is not None:
            # Divide by 2 because each sample has two associated haplotypes
            return len(self.__haplotypes) // 2
        else:
            # Divide by 2 because columns represent haplotypes
            return self.__lai.shape[1] // 2

    @property
    def n_ancestries(self) -> int:
        """
        Retrieve `n_ancestries`.

        Returns:
            **int:** The total number of unique ancestries.
        """
        return len(np.unique(self.__lai))
    
    @property
    def n_haplotypes(self) -> int:
        """
        Retrieve `n_haplotypes`.

        Returns:
            **int:** The total number of haplotypes.
        """
        if self.__haplotypes is not None:
            return len(self.__haplotypes)
        else:
            return self.__lai.shape[1]

    @property
    def n_windows(self) -> int:
        """
        Retrieve `n_windows`.

        Returns:
            **int:** The total number of genomic windows.
        """
        return self.__lai.shape[0]

    def copy(self) -> 'LocalAncestryObject':
        """
        Create and return a copy of `self`.

        Returns:
            **LocalAncestryObject:** 
                A new instance of the current object.
        """
        return copy.copy(self)

    def keys(self) -> List[str]:
        """
        Retrieve a list of public attribute names for `self`.

        Returns:
            **list of str:** 
                A list of attribute names, with internal name-mangling removed, 
                for easier reference to public attributes in the instance.
        """
        return [attr.replace('_LocalAncestryObject__', '').replace('_AncestryObject__', '') for attr in vars(self)]

    def filter_windows(
            self,
            indexes: Union[int, Sequence[int], np.ndarray],
            include: bool = True,
            inplace: bool = False
        ) -> Optional['LocalAncestryObject']:
        """
        Filter genomic windows based on specified indexes. 

        This method updates the `lai` attribute to include or exclude the specified genomic windows. 
        Attributes such as `window_sizes`, `centimorgan_pos`, `chromosomes`, and `physical_pos` will also be 
        updated accordingly if they are not None. The order of genomic windows is preserved.

        Negative indexes are supported and follow 
        [NumPy's indexing conventions](https://numpy.org/doc/stable/user/basics.indexing.html). 

        Args:
            indexes (int or array-like of int): 
                Index(es) of the windows to include or exclude. Can be a single integer or a
                sequence of integers. Negative indexes are supported.
            include (bool, default=True): 
                If True, includes only the specified windows. If False, excludes the specified
                windows. Default is True.
            inplace (bool, default=False): 
                If True, modifies `self` in place. If False, returns a new `LocalAncestryObject` with 
                the windows filtered. Default is False.

        Returns:
            **Optional[LocalAncestryObject]:** 
                A new `LocalAncestryObject` with the specified windows filtered if `inplace=False`. 
                If `inplace=True`, modifies `self` in place and returns None.
        """
        # Convert indexes to a NumPy array
        indexes = np.atleast_1d(indexes)

        # Get total number of windows
        n_windows = self.n_windows

        # Validate indexes, allowing negative indexes
        if np.any((indexes < -n_windows) | (indexes >= n_windows)):
            raise IndexError("One or more indexes are out of bounds.")

        # Create boolean mask
        mask = np.zeros(n_windows, dtype=bool)
        mask[indexes] = True

        # Invert mask if `include=False`
        if not include:
            mask = ~mask
        
        # Filter `lai`
        filtered_lai = self['lai'][mask, :] 
        
        # Filter `window_sizes`, `chromosomes`, `centimorgan_pos`, and `physical_pos`, checking if they are None before filtering
        filtered_window_sizes = self['window_sizes'][mask] if self['window_sizes'] is not None else None
        filtered_chromosomes = self['chromosomes'][mask] if self['chromosomes'] is not None else None
        filtered_centimorgan_pos = self['centimorgan_pos'][mask, :] if self['centimorgan_pos'] is not None else None
        filtered_physical_pos = self['physical_pos'][mask, :] if self['physical_pos'] is not None else None

        # Modify the original object if `inplace=True`, otherwise create and return a copy
        if inplace:
            self['lai'] = filtered_lai
            if filtered_window_sizes is not None:
                self['window_sizes'] = filtered_window_sizes
            if filtered_chromosomes is not None:
                self['chromosomes'] = filtered_chromosomes
            if filtered_centimorgan_pos is not None:
                self['centimorgan_pos'] = filtered_centimorgan_pos
            if filtered_physical_pos is not None:
                self['physical_pos'] = filtered_physical_pos
            return None
        else:
            laiobj = self.copy()
            laiobj['lai'] = filtered_lai
            if filtered_window_sizes is not None:
                laiobj['window_sizes'] = filtered_window_sizes
            if filtered_chromosomes is not None:
                laiobj['chromosomes'] = filtered_chromosomes
            if filtered_centimorgan_pos is not None:
                laiobj['centimorgan_pos'] = filtered_centimorgan_pos
            if filtered_physical_pos is not None:
                laiobj['physical_pos'] = filtered_physical_pos
            return laiobj

    def filter_samples(
        self,
        samples: Optional[Union[str, Sequence[str], np.ndarray, None]] = None,
        indexes: Optional[Union[int, Sequence[int], np.ndarray, None]] = None,
        include: bool = True,
        reorder: bool = False,
        inplace: bool = False
    ) -> Optional['LocalAncestryObject']:
        """
        Filter samples based on specified names or indexes.

        This method updates the `lai`, `haplotypes`, and `samples` attributes to include or exclude the specified 
        samples. Each sample is associated with two haplotypes, which are included or excluded together.
        The order of the samples is preserved. Set `reorder=True` to match the ordering of the
        provided `samples` and/or `indexes` lists when including.

        If both samples and indexes are provided, any sample matching either a name in samples or an index in 
        indexes will be included or excluded.
        
        Negative indexes are supported and follow 
        [NumPy's indexing conventions](https://numpy.org/doc/stable/user/basics.indexing.html).

        Args:
            samples (str or array_like of str, optional): 
                 Name(s) of the samples to include or exclude. Can be a single sample name or a
                 sequence of sample names. Default is None.
            indexes (int or array_like of int, optional):
                Index(es) of the samples to include or exclude. Can be a single index or a sequence
                of indexes. Negative indexes are supported. Default is None.
            include (bool, default=True): 
                If True, includes only the specified samples. If False, excludes the specified
                samples. Default is True.
            inplace (bool, default=False): 
                If True, modifies `self` in place. If False, returns a new `LocalAncestryObject` with the 
                samples filtered. Default is False.

        Returns:
            **Optional[LocalAncestryObject]:** 
                A new `LocalAncestryObject` with the specified samples filtered if `inplace=False`. 
                If `inplace=True`, modifies `self` in place and returns None.
        """
        if samples is None and indexes is None:
            raise UserWarning("At least one of 'samples' or 'indexes' must be provided.")

        n_haplotypes = self.n_haplotypes
        n_samples = self.n_samples

        # Create mask based on sample names
        if samples is not None:
            samples = np.asarray(samples).ravel()
            # Extract sample names from haplotype identifiers
            haplotype_ids = np.array(self['haplotypes'])
            sample_names = np.array([hap.split('.')[0] for hap in haplotype_ids])
            # Create mask for haplotypes belonging to specified samples
            mask_samples = np.isin(sample_names, samples)
        else:
            mask_samples = np.zeros(n_haplotypes, dtype=bool)

        # Create mask based on sample indexes
        if indexes is not None:
            indexes = np.asarray(indexes).ravel()

            # Validate indexes, allowing negative indexes
            out_of_bounds_indexes = indexes[(indexes < -n_samples) | (indexes >= n_samples)]
            if out_of_bounds_indexes.size > 0:
                raise ValueError(f"One or more sample indexes are out of bounds.")

            # Adjust negative indexes
            indexes = np.mod(indexes, n_samples)
            
            # Get haplotype indexes for the specified sample indexes
            haplotype_indexes = np.concatenate([2*indexes, 2*indexes+1])
            # Create mask for haplotypes
            mask_indexes = np.zeros(n_haplotypes, dtype=bool)
            mask_indexes[haplotype_indexes] = True
        else:
            mask_indexes = np.zeros(n_haplotypes, dtype=bool)

        # Combine masks using logical OR (union of samples)
        mask_combined = mask_samples | mask_indexes

        if not include:
            mask_combined = ~mask_combined

        # Optionally compute an ordering of selected samples that follows the provided lists
        ordered_sample_indices = None
        sample_mask = mask_combined.reshape(-1, 2).any(axis=1)
        if include and reorder:
            sel_sample_indices = np.where(sample_mask)[0]
            ordered_list: List[int] = []
            added = np.zeros(self.n_samples, dtype=bool)

            # Source of sample names for ordering logic
            haplotype_ids = np.array(self['haplotypes'])
            sample_names_by_sample = np.array([hap.split('.')[0] for hap in haplotype_ids])[::2]

            # Respect the order in `samples`
            if samples is not None:
                for s in np.atleast_1d(samples):
                    # Find the sample index by name (first occurrence)
                    matches = np.where(sample_names_by_sample == s)[0]
                    for idx in matches:
                        if sample_mask[idx] and not added[idx]:
                            ordered_list.append(int(idx))
                            added[idx] = True

            # Then respect the order in `indexes`
            if indexes is not None:
                adj_idx = np.mod(np.atleast_1d(indexes), self.n_samples)
                for idx in adj_idx:
                    if sample_mask[idx] and not added[idx]:
                        ordered_list.append(int(idx))
                        added[idx] = True

            # Append any remaining selected samples in their original order
            for idx in sel_sample_indices:
                if not added[idx]:
                    ordered_list.append(int(idx))

            ordered_sample_indices = np.asarray(ordered_list, dtype=int)

        # Filter / reorder arrays
        if ordered_sample_indices is not None:
            hap_idx = np.concatenate([2*ordered_sample_indices, 2*ordered_sample_indices + 1])
            filtered_lai = self['lai'][:, hap_idx]
            filtered_haplotypes = np.array(self['haplotypes'])[hap_idx].tolist()
            filtered_samples = (
                np.array(self['samples'])[ordered_sample_indices].tolist()
                if self['samples'] is not None else None
            )
        else:
            # Filter `lai`
            filtered_lai = self['lai'][:, mask_combined]
            # Filter `haplotypes`
            filtered_haplotypes = np.array(self['haplotypes'])[mask_combined].tolist()
            # Filter `samples`, checking if they are None before filtering
            filtered_samples = np.array(self['samples'])[sample_mask].tolist() if self['samples'] is not None else None

        if inplace:
            self['haplotypes'] = filtered_haplotypes
            self['samples'] = filtered_samples
            self['lai'] = filtered_lai
            return None
        else:
            laiobj = self.copy()
            laiobj['haplotypes'] = filtered_haplotypes
            laiobj['samples'] = filtered_samples
            laiobj['lai'] = filtered_lai
            return laiobj

    def convert_to_snp_level(
        self,
        snpobject: Optional['SNPObject'] = None,
        variants_chrom: Optional[np.ndarray] = None,
        variants_pos: Optional[np.ndarray] = None,
        variants_ref: Optional[np.ndarray] = None,
        variants_alt: Optional[np.ndarray] = None,
        variants_filter_pass: Optional[np.ndarray] = None,
        variants_id: Optional[np.ndarray] = None,
        variants_qual: Optional[np.ndarray] = None,
        lai_format: str = "3D"
    ) -> 'SNPObject':
        """
        Convert `self` into a `snputils.snp.genobj.SNPObject` SNP-level Local Ancestry Information (LAI), 
        with optional support for SNP data.
        
        If SNP positions (`variants_pos`) and/or chromosomes (`variants_chrom`) are not specified, the method generates 
        SNPs uniformly across the start and end positions of each genomic window. Otherwise, the provided SNP 
        coordinates are used to assign ancestry values based on their respective windows.

        If a `SNPObject` is provided, its attributes are used unless explicitly overridden by the function arguments.
        In that case, the SNPObject is updated with the (optional) new attributes and the computed `calldata_lai`, then returned.

        Args:
            snpobject (SNPObject, optional):
                An existing `SNPObject` to extract SNP attributes from.
            variants_chrom (array of shape (n_snps,), optional): 
                An array containing the chromosome for each SNP.
            variants_pos (array of shape (n_snps,), optional): 
                An array containing the chromosomal positions for each SNP.
            variants_ref (array of shape (n_snps,), optional): 
                An array containing the reference allele for each SNP.
            variants_alt (array of shape (n_snps,), optional): 
                An array containing the alternate allele for each SNP.
            variants_filter_pass (array of shape (n_snps,), optional): 
                An array indicating whether each SNP passed control checks.
            variants_id (array of shape (n_snps,), optional): 
                An array containing unique identifiers (IDs) for each SNP.
            variants_qual (array of shape (n_snps,), optional): 
                An array containing the Phred-scaled quality score for each SNP.
            lai_format (str, optional):
                Determines the shape of `calldata_lai`:
                    - `"3D"` (default): Shape `(n_snps, n_samples, 2)`.
                    - `"2D"`: Shape `(n_snps, n_samples * 2)`.

        Returns:
            **SNPObject**: 
                A `SNPObject` containing SNP-level ancestry data and updated SNP attributes.
        """
        from snputils.snp.genobj.snpobj import SNPObject

        assert lai_format in {"2D", "3D"}, "Invalid `lai_format`. Must be '2D' or '3D'."

        # Extract attributes from SNPObject if provided
        if snpobject is not None:
            variants_chrom = variants_chrom or snpobject.variants_chrom
            variants_pos = variants_pos or snpobject.variants_pos
            variants_ref = variants_ref or snpobject.variants_ref
            variants_alt = variants_alt or snpobject.variants_alt
            variants_filter_pass = variants_filter_pass or snpobject.variants_filter_pass
            variants_id = variants_id or snpobject.variants_id
            variants_qual = variants_qual or snpobject.variants_qual

        n_samples = self.n_samples
        lai_reshaped = self.lai.reshape(self.n_windows, n_samples, 2).astype(int) if lai_format == "3D" else None

        if variants_pos is None or variants_chrom is None:
            # Generate all SNP positions and corresponding chromosome labels between window boundaries
            variants_pos_list = []
            variants_chrom_list = []
            ancestry_list = []

            for i in range(self.n_windows):
                start = int(self.physical_pos[i, 0])
                end = int(self.physical_pos[i, 1])
                chrom = self.chromosomes[i]

                # Generate SNP positions at each base pair within the window range
                positions_in_window = np.arange(start, end + 1)
                if positions_in_window.size == 0:
                    continue  # Skip windows that contain no valid SNP positions

                n_positions = positions_in_window.size
                variants_pos_list.append(positions_in_window)
                variants_chrom_list.append(np.full(n_positions, chrom))

                ancestry_repeated = (
                    np.repeat(lai_reshaped[i, np.newaxis, :, :], n_positions, axis=0)
                    if lai_format == "3D" else np.repeat(self.lai[i, np.newaxis, :], n_positions, axis=0)
                )
                ancestry_list.append(ancestry_repeated)

            # Store SNP positions, their corresponding chromosome labels, and their associated ancestry
            variants_pos = np.concatenate(variants_pos_list)
            variants_chrom = np.concatenate(variants_chrom_list)
            calldata_lai = np.concatenate(ancestry_list)
        else:
            # Use the provided SNP positions and chromosomes
            n_snps = len(variants_pos)
            if len(variants_chrom) != n_snps:
                raise ValueError("`variants_pos` and `variants_chrom` must have the same length.")

            # Initialize an array to store the corresponding window index for each SNP
            # Default value is -1, meaning no matching window found
            snp_to_window_indices = np.full(n_snps, -1, dtype=int)

            # Identify unique chromosome names sorted by order of appearence
            _, idx = np.unique(variants_chrom, return_index=True)
            unique_chroms = variants_chrom[np.sort(idx)]

            # Iterate through each unique chromosome to map SNPs to windows
            for chrom in unique_chroms:
                # Get indices of SNPs that belong to the current chromosome
                snp_indices = np.where(variants_chrom == chrom)[0]
                snp_pos_chr = variants_pos[snp_indices]
                
                # Get indices of windows that belong to the current chromosome
                window_indices = np.where(self.chromosomes == chrom)[0]
                if window_indices.size == 0:
                    continue  # Skip if no windows exist for this chromosome
                
                # Extract start and end positions of the windows on this chromosome
                window_starts_chr = self.physical_pos[:, 0][window_indices]
                window_ends_chr = self.physical_pos[:, 1][window_indices]
                
                # Find the right-most window that a SNP would fit into (sorted order)
                inds = np.searchsorted(window_starts_chr, snp_pos_chr, side='right') - 1
                
                # Mask valid SNPs: ensure they are within a valid range and fall inside window boundaries
                valid_mask = (inds >= 0) & (inds < len(window_starts_chr)) & (snp_pos_chr <= window_ends_chr[inds])

                # Assign valid SNPs to their corresponding window indices
                snp_to_window_indices[snp_indices[valid_mask]] = window_indices[inds[valid_mask]]
                log.debug(f"Number of SNPs within window ranges for chromosome {chrom}: {valid_mask.sum()}")

            # Initialize SNP-level ancestry array with NaNs as default
            shape = (n_snps, n_samples, 2) if lai_format == "3D" else (n_snps, n_samples * 2)
            calldata_lai = np.full(shape, np.nan, dtype='uint8')

            # Assign ancestry values to SNPs with valid window assignments
            valid_mask = (snp_to_window_indices != -1)
            snp_indices = np.where(valid_mask)[0]
            snp_to_window_indices = snp_to_window_indices[snp_indices]

            if lai_format == "3D":
                calldata_lai[snp_indices] = lai_reshaped[snp_to_window_indices]
            else:  # "2D"
                calldata_lai[snp_indices] = self.lai[snp_to_window_indices]

        if snpobject is not None:
            # If a SNPObject was provided, update its attributes with any new values and add `calldata_lai``
            snpobject.variants_chrom = variants_chrom
            snpobject.variants_pos = variants_pos
            snpobject.variants_ref = variants_ref
            snpobject.variants_alt = variants_alt
            snpobject.variants_filter_pass = variants_filter_pass
            snpobject.variants_id = variants_id
            snpobject.variants_qual = variants_qual
            snpobject.calldata_lai = calldata_lai
            snpobject.ancestry_map = self.ancestry_map
            return snpobject
        else:
            # Otherwise, create a new SNPObject
            return SNPObject(
                calldata_lai=calldata_lai.view(),
                samples=self.samples,
                variants_ref=variants_ref.view() if isinstance(variants_ref, np.ndarray) else variants_ref,
                variants_alt=variants_alt.view() if isinstance(variants_alt, np.ndarray) else variants_alt,
                variants_filter_pass=variants_filter_pass.view() if isinstance(variants_filter_pass, np.ndarray) else variants_filter_pass,
                variants_chrom=variants_chrom.view(),
                variants_id=variants_id.view() if isinstance(variants_id, np.ndarray) else variants_id,
                variants_pos=variants_pos.view(),
                variants_qual=variants_qual.view() if isinstance(variants_qual, np.ndarray) else variants_qual,
                ancestry_map=self.ancestry_map
            )

    def _sanity_check(self) -> None:
        """
        Perform sanity checks on the parsed data to ensure data integrity.

        This method checks that all unique ancestries in LAI are represented 
        in the ancestry map.

        Args:
            lai (np.ndarray): The LAI data array.
            ancestry_map (dict, optional): A dictionary mapping ancestry codes to region names, if available.
        """
        # Get unique ancestries from LAI data
        unique_ancestries = np.unique(self.lai)

        if self.ancestry_map is not None:
            # Check if all unique ancestries in the LAI are present in the ancestry map
            for ancestry in unique_ancestries:
                if str(ancestry) not in self.ancestry_map:
                    warnings.warn(
                        f"Ancestry '{ancestry}' found in LAI data is not represented in the ancestry map."
                    )

    def save(self, file: Union[str, Path]) -> None:
        """
        Save the data stored in `self` to a specified file.
        If the file already exists, it will be overwritten.

        The format of the saved file is determined by the file extension provided in the `file` 
        argument.

        **Supported formats:**

        - `.msp`: Text-based MSP format.
        - `.msp.tsv`: Text-based MSP format with TSV extension.
        - `.pkl`: Pickle format for saving `self` in serialized form.

        Args:
            file (str or pathlib.Path): 
                Path to the file where the data will be saved. The extension of the file determines the save format. 
                Supported extensions: `.msp`, `.msp.tsv`, `.pkl`.
        """
        path = Path(file)
        suffixes = [suffix.lower() for suffix in path.suffixes]

        if suffixes[-2:] == ['.msp', '.tsv'] or suffixes[-1] == '.msp':
            self.save_msp(file)
        elif suffixes[-1] == '.pkl':
            self.save_pickle(file)
        else:
            raise ValueError(
                f"Unsupported file extension: {suffixes[-1]}"
                "Supported extensions are: .msp, .msp.tsv, .pkl."
            )

    def save_msp(self, file: Union[str, Path]) -> None:
        """
        Save the data stored in `self` to a `.msp` file.
        If the file already exists, it will be overwritten.

        Args:
            file (str or pathlib.Path): 
                Path to the file where the data will be saved. It should end with `.msp` or `.msp.tsv`. 
                If the provided path does not have one of these extensions, the `.msp` extension will be appended.
        """
        from snputils.ancestry.io.local.write import MSPWriter

        MSPWriter(self, file).write()

    def save_pickle(self, file: Union[str, Path]) -> None:
        """
        Save `self` in serialized form to a `.pkl` file.
        If the file already exists, it will be overwritten.

        Args:
            file (str or pathlib.Path): 
                Path to the file where the data will be saved. It should end with `.pkl`. 
                If the provided path does not have this extension, it will be appended.
        """
        import pickle
        with open(file, 'wb') as file:
            pickle.dump(self, file)



import numpy as np

from snputils.snp.io.read import SNPReader


def test_auto_reader(data_path, snpobj_pgen):
    reader = SNPReader(data_path + "/pgen/subset.pgen")
    snpobj = reader.read(sum_strands=False)

    assert np.array_equal(snpobj.calldata_gt, snpobj_pgen.calldata_gt)



import platform
import zipfile
import subprocess
from pathlib import Path

from snputils._utils.data_home import get_data_home
from snputils._utils.download import download_url


def download_plink(suppress_output: bool = True) -> Path:
    data_path = get_data_home()
    plink_path = data_path / "plink2"

    if not plink_path.exists():
        system = platform.system()
        is_arm = platform.machine().lower().startswith(('arm', 'aarch'))

        plink_urls = {
            ("Darwin", True): "plink2_mac_arm64_20241114.zip",  # macOS ARM
            ("Linux", False): "plink2_linux_x86_64_20241114.zip",  # Linux x86_64
        }

        try:
            zip_filename = plink_urls[(system, is_arm)]
        except KeyError:
            raise RuntimeError(f"Unsupported platform: {system} {platform.machine()}")

        plink_url = f"https://s3.amazonaws.com/plink2-assets/alpha6/{zip_filename}"

        zip_path = data_path / zip_filename

        download_url(plink_url, zip_path, show_progress=not suppress_output)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(data_path)
        subprocess.run(["chmod", "+x", plink_path], cwd=str(data_path),
                       stdout=subprocess.DEVNULL if suppress_output else None)

    return plink_path


def execute_plink_cmd(args, cwd=None, suppress_output=True) -> subprocess.CompletedProcess:
    cwd = cwd or get_data_home()
    plink_path = download_plink(suppress_output=suppress_output)  # Download plink if it does not exist
    args = [arg for arg in args if arg is not None]  # Remove None arguments
    return subprocess.run([str(plink_path), *args], cwd=str(cwd),
                          stdout=subprocess.DEVNULL if suppress_output else None)



import argparse
from typing import List


def parse_dummy_tool_args(argv):
    parser = argparse.ArgumentParser(prog='dummy_tool', description='Dummy tool as an example.')
    parser.add_argument('--what_to_print', type=str, default="Dummy tool.", help='Sentence to print.')
    return parser.parse_args(argv)


def dummy_tool(argv: List[str]):
    args = parse_dummy_tool_args(argv)
    print(args.what_to_print)
    return 0



import abc
from pathlib import Path
from typing import Union

from snputils.ancestry.genobj.wide import GlobalAncestryObject


class WideBaseWriter(abc.ABC):
    """
    Abstract class for global ancestry writers.
    """ 
    def __init__(
        self, 
        wideobj: GlobalAncestryObject, 
        file_prefix: Union[str, Path]
    ) -> None:
        """
        Args:
            wideobj (GlobalAncestryObject): 
                A GlobalAncestryObject instance.
            file_prefix (str or pathlib.Path): 
                Prefix for output file names, including directory path but excluding file extensions. 
                The prefix is used to generate specific file names for each output, with file-specific 
                suffixes appended as described above (e.g., `file_prefix.n_ancestries.Q` for the Q matrix file).
        """
        self.__wideobj = wideobj
        self.__file_prefix = Path(file_prefix)

    @property
    def wideobj(self) -> GlobalAncestryObject:
        """
        Retrieve `wideobj`.

        Returns:
            GlobalAncestryObject: A GlobalAncestryObject instance.
        """
        return self.__wideobj

    @property
    def file_prefix(self) -> Path:
        """
        Retrieve `file_prefix`.

        Returns:
            pathlib.Path: 
                Prefix for output file names, including directory path but excluding file extensions. 
                The prefix is used to generate specific file names for each output, with file-specific 
                suffixes appended as described above (e.g., `file_prefix.n_ancestries.Q` for the Q matrix file).
        """
        return self.__file_prefix

    @abc.abstractmethod
    def write(self) -> None:
        """
        Abstract method to write wide ancestry files.

        Subclasses must implement this method to write the data from the Q and P files, 
        as well as sample identifiers, SNPs, and ancestries if available.
        """
        pass



import pathlib
from typing import Union

from snputils.snp.genobj import SNPObject


def read_snp(filename: Union[str, pathlib.Path], **kwargs) -> SNPObject:
    """
    Automatically detect the file format and read it into a SNPObject.

    Args:
        filename: Filename of the file to read.
        **kwargs: Additional arguments passed to the reader method.

    Raises:
        ValueError: If the filename does not have an extension or the extension is not supported.
    """
    from snputils.snp.io.read.auto import SNPReader

    return SNPReader(filename).read(**kwargs)


def read_bed(filename: Union[str, pathlib.Path], **kwargs) -> SNPObject:
    """
    Read a BED fileset into a SNPObject.

    Args:
        filename: Filename of the BED fileset to read.
        **kwargs: Additional arguments passed to the reader method. See :class:`snputils.snp.io.read.bed.BEDReader` for possible parameters.
    """
    from snputils.snp.io.read.bed import BEDReader

    return BEDReader(filename).read(**kwargs)


def read_pgen(filename: Union[str, pathlib.Path], **kwargs) -> SNPObject:
    """
    Read a PGEN fileset into a SNPObject.

    Args:
        filename: Filename of the PGEN fileset to read.
        **kwargs: Additional arguments passed to the reader method. See :class:`snputils.snp.io.read.pgen.PGENReader` for possible parameters.
    """
    from snputils.snp.io.read.pgen import PGENReader

    return PGENReader(filename).read(**kwargs)


def read_vcf(filename: Union[str, pathlib.Path], 
             backend: str = 'polars',
             **kwargs) -> SNPObject:
    """
    Read a VCF fileset into a SNPObject.

    Args:
        filename: Filename of the VCF fileset to read.
        backend: Backend to use for reading the VCF file. Options are 'polars' or 'scikit-allel'.
        **kwargs: Additional arguments passed to the reader method. See :class:`snputils.snp.io.read.vcf.VCFReader` for possible parameters.
    """
    from snputils.snp.io.read.vcf import VCFReader, VCFReaderPolars
    if backend == 'polars':
        print(f"Reading {filename} with polars backend")
        return VCFReaderPolars(filename).read(**kwargs)
    else:
        print(f"Reading {filename} with scikit-allel backend")
        return VCFReader(filename).read(**kwargs)



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("snputils")
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"

from .snp import SNPObject, SNPReader, BEDReader, PGENReader, VCFReader, BEDWriter, PGENWriter, VCFWriter, read_snp, read_bed, read_pgen, read_vcf
from .ancestry import LocalAncestryObject, GlobalAncestryObject, MSPReader, MSPWriter, AdmixtureMappingVCFWriter, AdmixtureReader, AdmixtureWriter, read_lai, read_msp, read_adm, read_admixture
from .phenotype import MultiPhenotypeObject, UKBPhenotypeObject, MultiPhenTabularReader, UKBPhenReader
from . import visualization as viz
from .datasets import load_dataset



import logging
import joblib
from typing import Union
from pathlib import Path

from snputils.snp.genobj import SNPObject

log = logging.getLogger(__name__)


class VCFWriter:
    """
    A writer class for exporting SNP data from a `snputils.snp.genobj.SNPObject` 
    into an `.vcf` file.
    """
    def __init__(self, snpobj: SNPObject, filename: str, n_jobs: int = -1, phased: bool = False):
        """
        Args:
            snpobj (SNPObject):
                A SNPObject instance.
            file (str or pathlib.Path): 
                Path to the file where the data will be saved. It should end with `.vcf`. 
                If the provided path does not have this extension, the `.vcf` extension will be appended.
            n_jobs: 
                Number of jobs to run in parallel. 
                - `None`: use 1 job unless within a `joblib.parallel_backend` context.  
                - `-1`: use all available processors.  
                - Any other integer: use the specified number of jobs.
            phased: 
                If True, genotype data is written in "maternal|paternal" format.  
                If False, genotype data is written in "maternal/paternal" format.
        """
        self.__snpobj = snpobj
        self.__filename = Path(filename)
        self.__n_jobs = n_jobs
        self.__phased = phased

    def write(
            self, 
            chrom_partition: bool = False,
            rename_missing_values: bool = True, 
            before: Union[int, float, str] = -1, 
            after: Union[int, float, str] = '.'
        ):
        """
        Writes the SNP data to VCF file(s).

        Args:
            chrom_partition (bool, optional):
                If True, individual VCF files are generated for each chromosome.
                If False, a single VCF file containing data for all chromosomes is created. Defaults to False.
            rename_missing_values (bool, optional):
                If True, renames potential missing values in `snpobj.calldata_gt` before writing. 
                Defaults to True.
            before (int, float, or str, default=-1): 
                The current representation of missing values in `calldata_gt`. Common values might be -1, '.', or NaN.
                Default is -1.
            after (int, float, or str, default='.'): 
                The value that will replace `before`. Default is '.'.
        """
        self.__chrom_partition = chrom_partition

        file_extensions = (".vcf", ".bcf")
        if self.__filename.suffix in file_extensions:
            self.__file_extension = self.__filename.suffix
            self.__filename = self.__filename.with_suffix('')
        else:
            self.__file_extension = ".vcf"

        # Optionally rename potential missing values in `snpobj.calldata_gt` before writing
        if rename_missing_values:
            self.__snpobj.rename_missings(before=before, after=after, inplace=True)

        data = self.__snpobj

        if self.__chrom_partition:
            chroms = data.unique_chrom

            for chrom in chroms:
                # Filter to include the data for the chromosome in particular
                data_chrom = data.filter_variants(chrom=chrom, inplace=False)

                log.debug(f'Storing chromosome {chrom}')
                self._write_chromosome_data(chrom, data_chrom)
        else:
            self._write_chromosome_data("All", data)

    def _write_chromosome_data(self, chrom, data_chrom):
        """
        Writes the SNP data for a specific chromosome to a VCF file.

        Args:
            chrom: The chromosome name.
            data_chrom: The SNPObject instance containing the data for the chromosome.
        """
        # Obtain npy matrix with SNPs
        npy3 = data_chrom.calldata_gt  # shape: (n_windows, n_samples, 2)
        n_windows, n_samples, _ = npy3.shape

        # Keep sample names if appropriate
        data_samples = data_chrom.samples if len(data_chrom.samples) == n_samples else [get_name() for _ in range(n_samples)]

        # Format output file
        if chrom == "All":
            file = self.__filename.with_suffix(self.__file_extension)
        else:
            file = self.__filename.parent / f"{self.__filename.stem}_{chrom}{self.__file_extension}"

        # Write VCF file
        out = open(self.__filename.with_suffix(self.__file_extension), "w")
        # --- write VCF header ---
        out.write("##fileformat=VCFv4.1\n")
        out.write('##FORMAT=<ID=GT,Number=1,Type=String,Description="Phased Genotype">\n')
        
        for c in set(data_chrom.variants_chrom):
            out.write(f"##contig=<ID={c}>\n")
        cols = ["#CHROM","POS","ID","REF","ALT","QUAL","FILTER","INFO","FORMAT"] + list(data_chrom.samples)
        out.write("\t".join(cols) + "\n")
        
        sep = "|" if self.__phased else "/"
        for i in range(n_windows):
            chrom = data_chrom.variants_chrom[i]
            pos = data_chrom.variants_pos[i]
            vid = data_chrom.variants_id[i]
            ref = data_chrom.variants_ref[i]
            alt = data_chrom.variants_alt[i]
        
            # build genotype list per sample, one small array at a time
            row = npy3[i]  # shape: (n_samples, 2)
            genotypes = [
                f"{row[s,0]}{sep}{row[s,1]}"
                for s in range(n_samples)
            ]
        
            line = "\t".join([
                str(chrom), str(pos), vid, ref, alt,
                ".", "PASS", ".", "GT", *genotypes
            ])
            out.write(line + "\n")
        
        out.close()


def process_genotype(npy, i, n_snps, phased):
    """
    Process the genotype data for a particular individual in "maternal|paternal" 
    format for each SNP.

    Args:
        npy: Array containing genotype data for multiple individuals.
        i: Index of the individual to process.
        n_snps: Number of SNPs.

    Returns:
        **genotype**: List with "maternal|paternal" for each SNP.
    """
    # Get the genotype data for the specified individual's maternal and paternal SNPs
    maternal = npy[i*2, :].astype(str)     # maternal strand is the even row
    paternal = npy[i*2 + 1, :].astype(str)  # paternal strand is the odd row

    # Create a list with "maternal|paternal" format for each SNP
    sep = "|" if phased else "/"
    lst = [maternal, [sep] * n_snps, paternal]
    genotype = list(map(''.join, zip(*lst)))

    return genotype



import urllib.request
from tqdm import tqdm


class DownloadProgressBar(tqdm):
    """
    tqdm progress bar for download progress.
    """
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        Update the progress bar.

        Args:
            b (int): Number of blocks transferred.
            bsize (int): Size of each block.
            tsize (int): Total size of the file.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path, show_progress=False):
    """
    Download a file from a URL.

    Args:
        url (str): URL of the file to download.
        output_path (str): Path to save the downloaded file.
        show_progress (bool): Whether to show download progress with a tqdm progress bar.
    """
    if show_progress:
        with DownloadProgressBar(unit='B', unit_scale=True,
                                 miniters=1, desc=url.split('/')[-1]) as t:
            urllib.request.urlretrieve(url, filename=output_path,
                                       reporthook=t.update_to)
    else:
        urllib.request.urlretrieve(url, filename=output_path)



from .local import LAIReader, MSPReader, MSPWriter, AdmixtureMappingVCFWriter, read_lai, read_msp
from .wide import AdmixtureReader, AdmixtureWriter, read_adm, read_admixture

__all__ = ['read_adm', 'read_admixture', 'read_lai', 'read_msp',
           'AdmixtureReader', 'AdmixtureWriter', 'LAIReader', 'MSPReader', 'MSPWriter', 'AdmixtureMappingVCFWriter']



import logging
import numpy as np
from pathlib import Path
from typing import Union, Optional

log = logging.getLogger(__name__)

from .base import WideBaseReader
from snputils.ancestry.genobj.wide import GlobalAncestryObject


class AdmixtureReader(WideBaseReader):
    """
    A reader class for parsing ADMIXTURE files and constructing a `snputils.ancestry.genobj.GlobalAncestryObject`.
    """
    def __init__(
        self,
        Q_file: Union[str, Path],
        P_file: Optional[Union[str, Path]] = None,
        sample_file: Optional[Union[str, Path]] = None,
        snp_file: Optional[Union[str, Path]] = None,
        ancestry_file: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Args:
            Q_file (str or pathlib.Path):
                Path to the file containing the Q matrix (per-sample ancestry proportions).
                It should end with .Q or .txt.
                The file should use space (' ') as the delimiter.
            P_file (str or pathlib.Path, optional):
                Path to the file containing the P/F matrix (per-ancestry SNP frequencies).
                It should end with .P or .txt.
                The file should use space (' ') as the delimiter. If None, P is not loaded.
            sample_file (str or pathlib.Path, optional):
                Path to the single-column file containing sample identifiers. 
                It should end with .fam or .txt.
                If None, sample identifiers are not loaded.
            snp_file (str or pathlib.Path, optional):
                Path to the single-column file containing SNP identifiers. 
                It should end with .bim or .txt.
                If None, SNP identifiers are not loaded.
            ancestry_file (str or pathlib.Path, optional):
                Path to the single-column file containing ancestry labels for each sample.
                It should end with .map or .txt.
                If None, ancestries are not loaded.
        """
        self.__Q_file = Path(Q_file)
        self.__P_file = Path(P_file) if P_file is not None else None
        self.__sample_file = Path(sample_file) if sample_file is not None else None
        self.__snp_file = Path(snp_file) if snp_file is not None else None
        self.__ancestry_file = Path(ancestry_file) if ancestry_file is not None else None

    @property
    def Q_file(self) -> Path:
        """
        Retrieve Q_file.

        Returns:
            **pathlib.Path:** 
                Path to the file containing the Q matrix (per-sample ancestry proportions).
                It should end with .Q or .txt.
                The file should use space (' ') as the delimiter.
        """
        return self.__Q_file

    @property
    def P_file(self) -> Optional[Path]:
        """
        Retrieve P_file.

        Returns:
            **pathlib.Path or None:** 
                Path to the file containing the P/F matrix (per-ancestry SNP frequencies).
                It should end with .P or .txt.
                The file should use space (' ') as the delimiter. If None, P is not loaded.
        """
        return self.__P_file

    @property
    def sample_file(self) -> Optional[Path]:
        """
        Retrieve sample_file.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing sample identifiers. 
                It should end with .fam or .txt.
                If None, sample identifiers are not loaded.
        """
        return self.__sample_file
    
    @property
    def snp_file(self) -> Optional[Path]:
        """
        Retrieve snp_file.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing SNP identifiers. 
                It should end with .bim or .txt.
                If None, SNP identifiers are not loaded.
        """
        return self.__snp_file

    @property
    def ancestry_file(self) -> Optional[Path]:
        """
        Retrieve ancestry_file.

        Returns:
            **pathlib.Path:** 
                Path to the single-column file containing ancestry labels for each sample.
                It should end with .map or .txt.
                If None, ancestries are not loaded.
        """
        return self.__ancestry_file

    def read(self) -> 'GlobalAncestryObject':
        """
        Read data from the provided ADMIXTURE files and construct a 
        snputils.ancestry.genobj.GlobalAncestryObject instance.

        **Expected ADMIXTURE files content:**

        - **Q_file**: 
            A text file containing the Q matrix with per-sample ancestry proportions. 
             Each row corresponds to a sample, and each column corresponds to an ancestry.
        - **P_file**: 
            A text file containing the P matrix with per-ancestry SNP frequencies.
            Each row corresponds to a SNP, and each column corresponds to an ancestry.

        Optional files (if provided):
        - **sample_file**: A single-column text file containing sample identifiers in order.
        - **snp_file**: A single-column text file containing SNP identifiers in order.
        - **ancestry_file**: A single-column text file containing ancestry labels for each sample.

        Returns:
            **GlobalAncestryObject:** 
                A GlobalAncestryObject instance.
        """
        log.info(f"Reading Q matrix from '{self.Q_file}'...")
        Q_mat = np.genfromtxt(self.Q_file, delimiter=' ')
        if self.P_file is not None:
            log.info(f"Reading P matrix from '{self.P_file}'...")
            P_mat = np.genfromtxt(self.P_file, delimiter=' ')
        else:
            P_mat = None

        samples = self._read_sample_ids()
        snps = self._read_snps()
        ancestries = self._read_ancestries()

        return GlobalAncestryObject(
            Q_mat,
            P_mat,
            samples=samples,
            snps=snps,
            ancestries=ancestries
        )

WideBaseReader.register(AdmixtureReader)



from .read import SNPReader, BEDReader, PGENReader, VCFReader, read_snp, read_bed, read_pgen, read_vcf
from .write import BEDWriter, PGENWriter, VCFWriter

__all__ = ['read_snp', 'read_bed', 'read_pgen', 'read_vcf',
           'SNPReader', 'BEDReader', 'PGENReader', 'VCFReader',
           'BEDWriter', 'PGENWriter', 'VCFWriter']



import logging
import pandas as pd
from pathlib import Path
from typing import Union

log = logging.getLogger(__name__)

from .base import PhenotypeBaseReader
from snputils.phenotype.genobj import UKBPhenotypeObject


class UKBPhenReader(PhenotypeBaseReader):
    """
    A class for reading data from a `.phe` file and constructing a `UKBPhenotypeObject`.
    """
    def __init__(self, file: Union[str, Path]) -> None:
        """
        Args:
            file (str or pathlib.Path): 
                Path to the `.phe` file containing UKB phenotype data.
        """
        self._file = file

    @property
    def file(self) -> Path:
        """
        Retrieve `file`.

        Returns:
            pathlib.Path: 
                Path to the `.phe` file containing UKB phenotype data.
        """
        return self.__file
    
    def read(self) -> 'UKBPhenotypeObject':
        """
        Read data from a `.phe ` file and construct a `UKBPhenotypeObject`.

        Returns:
            UKBPhenotypeObject: 
                A UKB phenotype object instance.
        """
        log.info(f"Reading .phe file from '{self.file}'...")
        
        # Load the `.phe` file with specified column names and data types
        phen_df = pd.read_csv(
            self.file,
            header=None,
            delim_whitespace=True,
            names=["FID", "IID", "status"],
            dtype={"FID": str, "IID": str, "status": int},
        )
        
        # Extract sample IDs for cases based on status
        cases_IDs = list(phen_df[phen_df["status"] == 2]["FID"])
        n_cases = len(cases_IDs)
        if n_cases == 0:
            raise ValueError("No case data available.")
        
        # Extract sample IDs for controls based on status
        controls_IDs = list(phen_df[phen_df["status"] == 1]["FID"])
        n_controls = len(controls_IDs)
        if n_controls == 0:
            raise ValueError("No control data available.")
        
        # Verify the sample count integrity
        sample_IDs = phen_df["FID"].tolist()
        n_samples = len(sample_IDs)
        if n_samples != (n_cases + n_controls):
            raise ValueError(
                "Total sample count does not match the combined count of cases and controls. "
                f"Expected {n_cases + n_controls}; found {n_samples}."
            )
        
        # Generate haplotypes for cases, controls, and all samples
        cases_haplotypes = [f"{case}.0" for case in cases_IDs] + [f"{case}.1" for case in cases_IDs]
        controls_haplotypes = [f"{control}.0" for control in controls_IDs] + [f"{control}.1" for control in controls_IDs]
        all_haplotypes = [f"{sample}.0" for sample in sample_IDs] + [f"{sample}.1" for sample in sample_IDs]
        
        return UKBPhenotypeObject(samples = sample_IDs,
                                  n_samples = n_samples,
                                  cases = cases_IDs,
                                  n_cases = n_cases,
                                  controls = controls_IDs,
                                  n_controls = n_controls,
                                  all_haplotypes = all_haplotypes,
                                  cases_haplotypes = cases_haplotypes,
                                  controls_haplotypes = controls_haplotypes
                                 )

PhenotypeBaseReader.register(UKBPhenReader)



import warnings
import torch
import numpy as np
import copy
from typing import Tuple, Optional, Union, List
from sklearn.decomposition import PCA as skPCA 

from snputils.snp.genobj.snpobj import SNPObject


def _svd_flip(u, v, u_based_decision=True):
    """
    Sign correction to ensure deterministic output from SVD.

    Adjusts the columns of u and the rows of v such that the loadings in the
    columns in u that are largest in absolute value are always positive.
    """
    if u_based_decision:
        # columns of u, rows of v
        max_abs_cols = torch.argmax(torch.abs(u), axis=0)
        signs = torch.sign(u[max_abs_cols, range(u.shape[1])])
        u *= signs
        v *= signs[:, None]
    else:
        # rows of v, columns of u
        max_abs_rows = torch.argmax(torch.abs(v), axis=1)
        signs = torch.sign(v[range(v.shape[0]), max_abs_rows])
        u *= signs
        v *= signs[:, None]

    return u, v


class TorchPCA:
    """
    A class for GPU-based Principal Component Analysis (PCA) with PyTorch tensors.
    
    This implementation leverages GPU acceleration to achieve significant performance improvements,
    being up to 25 times faster than `sklearn.decomposition.PCA` when running on a compatible GPU.
    """
    def __init__(self, n_components: int = 2, fitting: str = 'reduced'):
        """
        Args:
            n_components (int, default=2): 
                The number of principal components. If None, defaults to the minimum of `n_samples` and `n_snps`.
            fitting (str, default='reduced'): 
                The fitting approach for the SVD computation. Options:
                - `'full'`: Full Singular Value Decomposition (SVD).
                - `'reduced'`: Economy SVD.
                - `'lowrank'`: Low-rank approximation, which provides a faster but approximate solution.
                Default is 'reduced'.
        """
        self.__n_components = n_components
        self.__fitting = fitting
        self.__n_components_ = None
        self.__components_ = None
        self.__mean_ = None
        self.__X_new_ = None  # Store transformed SNP data

    def __getitem__(self, key):
        """
        To access an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    def __setitem__(self, key, value):
        """
        To set an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    @property
    def n_components(self) -> Optional[int]:
        """
        Retrieve `n_components`.

        Returns:
            **int:** The number of principal components.
        """
        return self.__n_components
    
    @n_components.setter
    def n_components(self, x: Optional[int]) -> None:
        """
        Update `n_components`.
        """
        if x is not None and (not isinstance(x, int) or x <= 0):
            raise ValueError("n_components must be a positive integer or None.")
        self.__n_components = x

    @property
    def fitting(self) -> str:
        """
        Retrieve `fitting`.

        Returns:
            **str:** 
                The fitting approach for the SVD computation.
        """
        return self.__fitting

    @fitting.setter
    def fitting(self, x: str) -> None:
        """
        Update `fitting`.
        """
        allowed = {"full", "reduced", "lowrank"}
        if x not in allowed:
            raise ValueError(f"fitting must be one of {sorted(allowed)}")
        self.__fitting = x

    @property
    def n_components_(self) -> Optional[int]:
        """
        Retrieve `n_components_`.
        
        Returns:
            **int:** 
                The effective number of components retained after fitting, 
                calculated as `min(self.n_components, min(n_samples, n_snps))`.
        """
        return self.__n_components_

    @n_components_.setter
    def n_components_(self, x: int) -> None:
        """
        Update `n_components_`.
        """
        self.__n_components_ = x

    @property
    def components_(self) -> Optional[torch.Tensor]:
        """
        Retrieve `components_`.
        
        Returns:
            **tensor of shape (n_components_, n_snps):** 
                Matrix of principal components, where each row is a principal component vector.
        """
        return self.__components_

    @components_.setter
    def components_(self, x: torch.Tensor) -> None:
        """
        Update `components_`.
        """
        self.__components_ = x

    @property
    def mean_(self) -> Optional[torch.Tensor]:
        """
        Retrieve `mean_`.

        Returns:
            **tensor of shape (n_snps,):** 
                Per-feature mean vector of the input data used for centering.
        """
        return self.__mean_

    @mean_.setter
    def mean_(self, x: torch.Tensor) -> None:
        """
        Update `mean_`.
        """
        self.__mean_ = x

    @property
    def X_new_(self) -> Optional[Union[torch.Tensor, np.ndarray]]:
        """
        Retrieve `X_new_`.

        Returns:
            **tensor of shape (n_samples, n_components_):** 
                The transformed SNP data projected onto the `n_components_` principal components.
        """
        return self.__X_new_

    @X_new_.setter
    def X_new_(self, x: torch.Tensor) -> None:
        """
        Update `X_new_`.
        """
        self.__X_new_ = x

    def copy(self) -> 'TorchPCA':
        """
        Create and return a copy of `self`.

        Returns:
            **TorchPCA:** 
                A new instance of the current object.
        """
        return copy.copy(self)

    def _fit(self, X: torch.Tensor) -> Tuple:
        """
        Internal method to fit the PCA model to the data `X`.

        Args:
            X (tensor of shape (n_samples, n_snps)): 
                Input SNP data used for fitting the model.

        Returns:
            Tuple: U, S, and Vt matrices from the SVD, where:
                - `U` has shape (n_samples, n_components).
                - `S` contains singular values (n_components).
                - `Vt` has shape (n_components, n_snps) and represents the principal components.
        """
        n_samples, n_snps = X.shape
        self.n_components_ = min(self.n_components or min(X.shape), min(X.shape))

        if self.n_components_ > min(X.shape):
            raise ValueError(f"n_components should be <= min(n_samples: {n_samples}, n_snps: {n_snps})")

        # Compute the mean to center the data
        self.mean_ = torch.mean(X, dim=0)

        # Choose SVD method based on the fitting type
        if self.fitting == "full":
            U, S, Vt = torch.linalg.svd(X - self.mean_, full_matrices=True)
        elif self.fitting == "reduced":
            U, S, Vt = torch.linalg.svd(X - self.mean_, full_matrices=False)
        elif self.fitting == "lowrank":
            U, S, V = torch.svd_lowrank(X, q=self.n_components_, M=self.mean_)
            Vt = V.mT  # Transpose V to get Vt
        else:
            raise ValueError(f"Unrecognized fitting method: {self.fitting}")

        # Select the first `n_components` columns and singular values
        U = U[:, :self.n_components_]
        S = S[:self.n_components_]
        Vt = Vt[:self.n_components_]

        # Ensure deterministic output for U and Vt signs
        U, Vt = _svd_flip(U, Vt)
        self.components_ = Vt

        return U, S, Vt

    def fit(self, X: torch.Tensor) -> 'TorchPCA':
        """
        Fit the model to the input SNP data.

        Args:
            X (tensor of shape (n_samples, n_snps)): 
                The SNP data matrix to fit the model.

        Returns:
            **TorchPCA:** 
                The fitted instance of `self`.
        """
        self._fit(X)
        return self

    def transform(self, X: torch.Tensor) -> torch.Tensor:
        """
        Apply dimensionality reduction to the input SNP data using the fitted model.

        Args:
            X (tensor of shape (n_samples, n_snps)): 
                The SNP data matrix to be transformed.

        Returns:
            **tensor of shape (n_samples, n_components_):** 
                The transformed SNP data projected onto the `n_components_` principal components, 
                stored in `self.X_new_`.
        """
        if self.components_ is None or self.mean_ is None:
            raise ValueError("The PCA model must be fitted before calling `transform`.")
        
        self.X_new_ = torch.matmul(X - self.mean_, self.components_.T)
        return self.X_new_

    def fit_transform(self, X):
        """
        Fit the model to the SNP data and apply dimensionality reduction on the same SNP data.

        Args:
            X (tensor of shape n_samples, n_snps): 
                The SNP data matrix used for both fitting and transformation.

        Returns:
            **tensor of shape (n_samples, n_components_):** 
                The transformed SNP data projected onto the `n_components_` principal components, 
                stored in `self.X_new_`.
        """
        U, S, _ = self._fit(X)
        self.X_new_ = U * S.unsqueeze(0)
        return self.X_new_


class PCA:
    """
    A class for Principal Component Analysis (PCA) on SNP data stored in a `snputils.snp.genobj.SNPObject`. 
    This class employs either 
    [sklearn.decomposition.PCA](https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html) 
    or the custom `TorchPCA` implementation for efficient GPU-accelerated analysis.
    
    The PCA class supports both separate and averaged strand processing for SNP data. If the `snpobj` parameter is 
    provided during instantiation, the `fit_transform` method will be automatically called, 
    applying PCA to transform the data according to the selected configuration.
    """
    def __init__(
        self, 
        snpobj: Optional['SNPObject'] = None, 
        backend: str = 'pytorch', 
        n_components: int = 2, 
        fitting: str = 'full', 
        device: str = 'cpu',
        average_strands: bool = True, 
        samples_subset: Optional[Union[int, List]] = None, 
        snps_subset: Optional[Union[int, List]] = None
    ):
        """
        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance.
            backend (str, default='pytorch'): 
                The backend to use (`'sklearn'` or `'pytorch'`). Default is 'pytorch'.
            n_components (int, default=2): 
                The number of principal components. Default is 2.
            fitting (str, default='full'): 
                The fitting approach to use for the SVD computation (only for `backend='pytorch'`). 
                - `'full'`: Full Singular Value Decomposition (SVD).
                - `'reduced'`: Economy SVD.
                - `'lowrank'`: Low-rank approximation, which provides a faster but approximate solution.
                Default is 'full'.
            device (str, default='cpu'): 
                Device to use (`'cpu'`, `'gpu'`, `'cuda'`, or `'cuda:<index>'`). Default is 'cpu'.
            average_strands (bool, default=True): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
            samples_subset (int or list of int, optional): 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
            snps_subset (int or list of int, optional): 
                Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
        """
        self.__snpobj = snpobj
        self.__backend = backend
        self.__n_components = n_components
        self.__fitting = fitting
        self.__device = self._process_device_argument(device)
        self.__average_strands = average_strands
        self.__samples_subset = samples_subset
        self.__snps_subset = snps_subset
        self.__X_ = None
        self.__X_new_ = None  # Store transformed SNP data
        self.__n_components_ = None
        self.__components_ = None
        self.__mean_ = None

        # Initialize PCA backend
        if self.backend == "pytorch":
            self.pca = TorchPCA(n_components=self.n_components, fitting=self.fitting)
        elif self.backend == "sklearn":
            self.pca = skPCA(n_components=self.n_components)
        else:
            raise ValueError("Unknown backend for PCA: ", backend)

        # Fit and transform if a `snpobj` is provided
        if self.snpobj is not None:
            self.fit_transform(snpobj)

    @property
    def snpobj(self) -> Optional['SNPObject']:
        """
        Retrieve `snpobj`.

        Returns:
            **SNPObject:** A SNPObject instance.
        """
        return self.__snpobj

    @snpobj.setter
    def snpobj(self, x) -> None:
        """
        Update `snpobj`.
        """
        self.__snpobj = x

    @property
    def backend(self) -> str:
        """
        Retrieve `backend`.
        
        Returns:
            **str:** The backend to use (`'sklearn'` or `'pytorch'`).
        """
        return self.__backend

    @backend.setter
    def backend(self, x: str) -> None:
        """
        Update `backend`.
        """
        self.__backend = x

    @property
    def n_components(self) -> int:
        """
        Retrieve `n_components`.
        
        Returns:
            **int:** The number of principal components.
        """
        return self.__n_components
    
    @n_components.setter
    def n_components(self, x: int) -> None:
        """
        Update `n_components`.
        """
        self.__n_components = x

    @property
    def fitting(self) -> str:
        """
        Retrieve `fitting`.

        Returns:
            **str:** 
                The fitting approach to use for the SVD computation (only for `backend='pytorch'`). 
                - `'full'`: Full Singular Value Decomposition (SVD).
                - `'lowrank'`: Low-rank approximation, which provides a faster but approximate solution.
        """
        return self.__fitting

    @fitting.setter
    def fitting(self, x: str) -> None:
        """
        Update `fitting`.
        """
        self.__fitting = x

    @property
    def device(self) -> torch.device:
        """
        Retrieve `device`.

        Returns:
            **torch.device:** Device to use (`'cpu'`, `'gpu'`, `'cuda'`, or `'cuda:<index>'`).
        """
        return self.__device

    @device.setter
    def device(self, x: str) -> None:
        """
        Update `device`.
        """
        self.__device = self._process_device_argument(x)

    @property
    def average_strands(self) -> bool:
        """
        Retrieve `average_strands`.

        Returns:
            **bool:** 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
        """
        return self.__average_strands

    @average_strands.setter
    def average_strands(self, x: bool) -> None:
        """
        Update `average_strands`.
        """
        self.__average_strands = x

    @property
    def samples_subset(self) -> Optional[Union[int, List[int]]]:
        """
        Retrieve `samples_subset`.

        Returns:
            **int or list of int:** 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
        """
        return self.__samples_subset

    @samples_subset.setter
    def samples_subset(self, x: Optional[Union[int, List[int]]]) -> None:
        """
        Update `samples_subset`.
        """
        self.__samples_subset = x

    @property
    def snps_subset(self) -> Optional[Union[int, List[int]]]:
        """
        Retrieve `snps_subset`.

        Returns:
            **int or list of int:** Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
        """
        return self.__snps_subset

    @snps_subset.setter
    def snps_subset(self, x: Optional[Union[int, List[int]]]) -> None:
        """
        Update `snps_subset`.
        """
        self.__snps_subset = x

    @property
    def n_components_(self) -> Optional[int]:
        """
        Retrieve `n_components_`.

        Returns:
            **int:** 
                The effective number of components retained after fitting, 
                calculated as `min(self.n_components, min(n_samples, n_snps))`.
        """
        return self.__n_components_

    @n_components_.setter
    def n_components_(self, x: int) -> None:
        """
        Update `n_components_`.
        """
        self.__n_components_ = x

    @property
    def components_(self) -> Optional[Union[torch.Tensor, np.ndarray]]:
        """
        Retrieve `components_`.

        Returns:
            **tensor or array of shape (n_components_, n_snps):** 
                Matrix of principal components, where each row is a principal component vector.
        """
        return self.__components_

    @components_.setter
    def components_(self, x: Union[torch.Tensor, np.ndarray]) -> None:
        """
        Update `components_`.
        """
        self.__components_ = x

    @property
    def mean_(self) -> Optional[Union[torch.Tensor, np.ndarray]]:
        """
        Retrieve `mean_`.

        Returns:
            **tensor or array of shape (n_snps,):** 
                Per-feature mean vector of the input data used for centering.
        """
        return self.__mean_

    @mean_.setter
    def mean_(self, x: Union[torch.Tensor, np.ndarray]) -> None:
        """
        Update `mean_`.
        """
        self.__mean_ = x

    @property
    def X_(self) -> Optional[Union[torch.Tensor, np.ndarray]]:
        """
        Retrieve `X_`.

        Returns:
            **tensor or array of shape (n_samples, n_snps):** 
                The SNP data matrix used to fit the model.
        """
        return self.__X_

    @X_.setter
    def X_(self, x: Union[torch.Tensor, np.ndarray]) -> None:
        """
        Update `X_`.
        """
        self.__X_ = x

    @property
    def X_new_(self) -> Optional[Union[torch.Tensor, np.ndarray]]:
        """
        Retrieve `X_new_`.

        Returns:
            **tensor or array of shape (n_samples, n_components_):** 
                The transformed SNP data projected onto the `n_components_` principal components.
        """
        return self.__X_new_

    @X_new_.setter
    def X_new_(self, x: Union[torch.Tensor, np.ndarray]) -> None:
        """
        Update `X_new_`.
        """
        self.__X_new_ = x

    def copy(self) -> 'PCA':
        """
        Create and return a copy of `self`.

        Returns:
            **PCA:** 
                A new instance of the current object.
        """
        return copy.copy(self)

    def _process_device_argument(self, device: str):
        """
        Process the device argument to map user-friendly device names to PyTorch device specifications.

        Args:
            device (str): Device specified by the user.

        Returns:
            torch.device: PyTorch device object.
        """
        if isinstance(device, str):
            device_lower = device.lower()
            if device_lower in ['cpu']:
                return torch.device('cpu')
            elif device_lower in ['gpu', 'cuda']:
                if torch.cuda.is_available():
                    return torch.device('cuda')
                else:
                    warnings.warn("CUDA is not available; using CPU instead.")
                    return torch.device('cpu')
            elif device_lower.startswith('cuda:'):
                if torch.cuda.is_available():
                    return torch.device(device_lower)
                else:
                    warnings.warn(f"CUDA is not available; requested device '{device}' is not available. Using CPU instead.")
                    return torch.device('cpu')
            else:
                raise ValueError(f"Unknown device type: '{device}'. Please use 'CPU', 'GPU', 'cuda', or 'cuda:<index>'.")
        elif isinstance(device, torch.device):
            return device
        else:
            raise TypeError(f"Device must be a string or torch.device, got {type(device)}.")

    def _get_data_from_snpobj(
            self, 
            snpobj: Optional['SNPObject'] = None, 
            average_strands: Optional[bool] = None, 
            samples_subset: Optional[Union[int, List]] = None, 
            snps_subset: Optional[Union[int, List]] = None
        ) -> Union[np.ndarray, torch.Tensor]:
        """
        Retrieve and prepare SNP data for PCA analysis, with options for selecting subsets and handling strands.

        This method processes SNP data stored in an `SNPObject`, which may include averaging of paternal 
        and maternal strands or selecting subsets of samples and SNPs. The prepared data is formatted 
        for use in PCA, optionally converted to a PyTorch tensor if the backend is set to 'pytorch'.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject object instance. If None, defaults to `self.snpobj`.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.
            samples_subset (int or list of int, optional): 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
                If None, defaults to `self.samples_subset`.
            snps_subset (int or list of int, optional): 
                Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
                If None, defaults to `self.snps_subset`.

            Returns:
                numpy.ndarray or torch.Tensor: 
                    The processed SNP data. If `backend` is set to 'pytorch', the data is returned as a 
                    PyTorch tensor on the specified `device`.
        """
        if snpobj is None:
            snpobj = self.snpobj
        if average_strands is None:
            average_strands = self.average_strands
        if samples_subset is None:
            samples_subset = self.samples_subset
        if snps_subset is None:
            snps_subset = self.snps_subset
            
        if snpobj.calldata_gt.ndim == 2:
            X = np.transpose(snpobj.calldata_gt.astype(float), (1,0))
        elif snpobj.calldata_gt.ndim == 3:
            X = np.transpose(snpobj.calldata_gt.astype(float), (1,0,2))
        
            if average_strands:
                X = np.mean(X, axis=2)
            else:
                X = np.reshape(X, (-1, X.shape[1]))
        else:
            raise ValueError(f"Invalid shape for `calldata_gt`: expected a 2D or 3D array, but got {snpobj.calldata_gt.ndim}D array.")
    
        # Handle sample and SNP subsets
        if isinstance(samples_subset, int):
            X = X[:samples_subset]
        elif isinstance(samples_subset, list):
            X = X[samples_subset]
        
        if isinstance(snps_subset, int):
            X = X[:, :snps_subset]
        elif isinstance(snps_subset, list):
            X = X[:, snps_subset]
        
        if self.backend == "pytorch":
            print(f"Converting data to PyTorch tensor on device {self.device}")
            X = torch.from_numpy(X).to(self.device)

        return X

    def fit(
            self, 
            snpobj: Optional['SNPObject'] = None, 
            average_strands: Optional[bool] = None, 
            samples_subset: Optional[Union[int, List]] = None, 
            snps_subset: Optional[Union[int, List]] = None
        ) -> 'PCA':
        """
        Fit the model to the input SNP data stored in the provided `snpobj`.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance. If None, defaults to `self.snpobj`.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.
            samples_subset (int or list of int, optional): 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
                If None, defaults to `self.samples_subset`.
            snps_subset (int or list of int, optional): 
                Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
                If None, defaults to `self.snps_subset`.

        Returns:
            **PCA:** 
                The fitted instance of `self`.
        """
        self.X_ = self._get_data_from_snpobj(snpobj, average_strands, samples_subset, snps_subset)
        self.pca.fit(self.X_)

        # Update attributes based on the fitted model
        self.n_components_ = self.pca.n_components_
        self.components_ = self.pca.components_
        self.mean_ = self.pca.mean_
        
        return self

    def transform(
            self, 
            snpobj: Optional['SNPObject'] = None, 
            average_strands: Optional[bool] = None, 
            samples_subset: Optional[Union[int, List]] = None, 
            snps_subset: Optional[Union[int, List]] = None
        ):
        """
        Apply dimensionality reduction to the input SNP data stored in the provided `snpobj` using the fitted model.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance. If None, defaults to `self.snpobj`.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.
            samples_subset (int or list of int, optional): 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
                If None, defaults to `self.samples_subset`.
            snps_subset (int or list of int, optional): 
                Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
                If None, defaults to `self.snps_subset`.

        Returns:
            **tensor or array of shape (n_samples, n_components):**
                The transformed SNP data projected onto the `n_components_` principal components,
                stored in `self.X_new_`.
        """
        # Retrieve or update the data to transform
        if snpobj is not None or self.X_ is None:
            self.X_ = self._get_data_from_snpobj(snpobj, average_strands, samples_subset, snps_subset)
        
        # Apply transformation using the fitted PCA model
        return self.pca.transform(self.X_)

    def fit_transform(self, snpobj: Optional['SNPObject'] = None, average_strands: Optional[bool] = None, 
                      samples_subset: Optional[Union[int, List]] = None, snps_subset: Optional[Union[int, List]] = None):
        """
        Fit the model to the SNP data stored in the provided `snpobj` and apply the dimensionality reduction 
        on the same SNP data.

        Args:
            snpobj (SNPObject, optional): 
                A SNPObject instance. If None, defaults to `self.snpobj`.
            average_strands (bool, optional): 
                True if the haplotypes from the two parents are to be combined (averaged) for each individual, or False otherwise.
                If None, defaults to `self.average_strands`.
            samples_subset (int or list of int, optional): 
                Subset of samples to include, as an integer for the first samples or a list of sample indices.
                If None, defaults to `self.samples_subset`.
            snps_subset (int or list of int, optional): 
                Subset of SNPs to include, as an integer for the first SNPs or a list of SNP indices.
                If None, defaults to `self.snps_subset`.

        Returns:
            **tensor or array of shape (n_samples, n_components):** 
                The transformed SNP data projected onto the `n_components_` principal components,
                stored in `self.X_new_`.
        """
        self.X_ = self._get_data_from_snpobj(snpobj, average_strands, samples_subset, snps_subset)
        self.X_new_ = self.pca.fit_transform(self.X_)

        # Update attributes based on the fitted model
        self.n_components_ = self.pca.n_components_
        self.components_ = self.pca.components_
        self.mean_ = self.pca.mean_

        return self.X_new_



from __future__ import annotations

import pathlib
from typing import Union


class SNPReader:
    def __new__(cls,
                filename: Union[str, pathlib.Path],
                vcf_backend: str = 'polars') -> SNPReader:
        """
        Automatically detect the SNP file format from the file extension, and return its corresponding reader.

        Args:
            filename: Filename of the file to read.
            vcf_backend: Backend to use for reading the VCF file. Options are 'polars' or 'scikit-allel'. Default is 'polars'.

        Raises:
            ValueError: If the filename does not have an extension or the extension is not supported.
        """
        filename = pathlib.Path(filename)
        suffixes = filename.suffixes
        if not suffixes:
            raise ValueError("The filename should have an extension when using SNPReader.")

        extension = suffixes[-2] if suffixes[-1].lower() in (".zst", ".gz") else suffixes[-1]
        extension = extension.lower()

        if extension == ".vcf":
            if vcf_backend == 'polars':
                from snputils.snp.io.read.vcf import VCFReaderPolars

                return VCFReaderPolars(filename)
            elif vcf_backend == 'scikit-allel':
                from snputils.snp.io.read.vcf import VCFReader

                return VCFReader(filename)
            else:
                raise ValueError(f"VCF backend not supported: {vcf_backend}")
        elif extension in (".bed", ".bim", ".fam"):
            from snputils.snp.io.read.bed import BEDReader

            return BEDReader(filename)
        elif extension in (".pgen", ".pvar", ".psam", ".pvar.zst"):
            from snputils.snp.io.read.pgen import PGENReader

            return PGENReader(filename)
        else:
            raise ValueError(f"File format not supported: {filename}")



import sys
import torch
import argparse
import logging
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Tuple, Union

from ..processing import PCA
from snputils.snp.io.read import VCFReader

log = logging.getLogger(__name__)


def parse_pca_args(argv):
    parser = argparse.ArgumentParser(prog='pca', description='Principal Component Analysis plot and save components.')

    # parser = argparse.ArgumentParser(description='.')

    parser.add_argument('--vcf_file', required=True, type=str, help='Load .vcf from this path.')
    parser.add_argument('--fig_path', required=True, type=str, help='Path used to save PCA plot.')
    parser.add_argument('--npy_path', required=True, type=str, help='Path used to save principal components in .npy format.') 
    parser.add_argument('--backend', required=True, type=str, default='sklearn', help='Backend used to perform PCA.') 
    
    
    return parser.parse_args(argv)


def plot_and_save_pca(argv: List[str]):
    args = parse_pca_args(argv)
    reader = VCFReader(args.vcf_file)
    snpobj = reader.read()

    pca = PCA(backend=args.backend, n_components=2)
    components = pca.fit_transform(snpobj)
    if args.backend == "pytorch":
        components = components.cpu().numpy()

    plt.figure(figsize=(10, 8))
    plt.scatter(components[:,0], components[:,1], linewidth=0, alpha=0.5)
    plt.xlabel("Principal Component 1", fontsize=20)
    plt.ylabel("Principal Component 2", fontsize=20)
    plt.tight_layout()

    # Save plot
    plt.savefig(args.fig_path)

    # Save components in .npy format    
    np.save(args.npy_path, components)
    return 0



import numpy as np
from typing import Optional, Tuple, Dict, cast
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as patches

from snputils.ancestry.genobj.local import LocalAncestryObject


def _custom_cmap(colors: Dict, padding: float = 1.05):
    """
    Create a custom colormap from a dictionary.

    Args:
        colors: Dictionary with levels as keys and color names as values.
        padding: Offset applied to levels. Defaults to 1.05.

    Returns:
        cmap: The custom colormap.
    """
    labels = sorted(float(key) for key in colors if key is not None)
    clrs = [colors[key] for key in labels]
    
    # Adjust levels to match the number of colors minus one
    levels = sorted([labels[0] - 1] + [x * padding for x in labels])

    cmap, _ = mcolors.from_levels_and_colors(levels, clrs)
    return cmap


def plot_lai(
    laiobj: LocalAncestryObject, 
    colors: Dict,
    sort: Optional[bool]=True,
    figsize: Optional[Tuple[float, float]]=None,
    legend: Optional[bool]=False,
    title: Optional[str]=None,
    fontsize: Optional[Dict[str, float]]=None,
    scale: int=2,
):
    """
    Plot LAI (Local Ancestry Inference) data with customizable options. Each row 
    represents the ancestry of a sample at the window level, distinguishing 
    between maternal and paternal strands. Whitespace is used to separate 
    individual samples.

    Args:
        laiobj: A LocalAncestryObject containing LAI data.
        colors: A dictionary with ancestry-color mapping.
        sort: If True, sort samples based on the most frequent ancestry. 
            Samples are displayed with the most predominant ancestry first, followed by the 
            second most predominant, and so on. Defaults to True.
        figsize: Figure size. If is None, the figure is displayed with a default size of 
            (25, 25). Defaults to None.
        legend: If True, display a legend. If ``sort==True``, ancestries in the 
            legend are sorted based on their total frequency in descending order. Defaults to False.
        title: Title for the plot. If None, no title is displayed. Defaults to None.
        fontsize: Font sizes for various plot elements. If None, default font sizes are used. 
            Defaults to None.
        scale: Number of times to duplicate rows for enhanced vertical visibility. Defaults to 2.
    """
    # The `lai` field is a 2D array containing the window-wise ancestry 
    # information for each individual. Consecutive rows of the transformed form 
    # correspond to the maternal and paternal ancestries of the same individual
    lai_T = laiobj.lai.T
    
    # Obtain number of samples and windows
    n_samples = int(lai_T.shape[0]/2)
    n_windows = lai_T.shape[1]
    
    if fontsize is None:
        fontsize = {
            'xticks' : 20, 
            'yticks' : 9, 
            'xlabel' : 20, 
            'ylabel' : 20,
            'legend' : 20
        }
    
    if sort:
        # Reshape `lai_T` to a 3D array where the third dimension represents 
        # pairs of consecutive rows, corresponding to maternal and paternal ancestries
        # Dimension: n_samples x 2 x n_windows
        maternal_paternal_pairs = lai_T.reshape(n_samples, 2, n_windows)
        
        # Reshape `maternal_paternal_pairs` to a 2D array where each row contains 
        # concatenated maternal and paternal ancestries
        # Dimension: n_samples x (2·n_windows)
        if maternal_paternal_pairs.ndim != 3:
            raise ValueError("maternal_paternal_pairs must be a 3D array, got array with ndim=" + str(maternal_paternal_pairs.ndim))
        num_samples, num_maternal_paternal, num_windows = cast(Tuple[int, int, int], maternal_paternal_pairs.shape)
        
        flat_ancestry_pairs = maternal_paternal_pairs.reshape(
            num_samples, num_maternal_paternal * num_windows
        )

        # Determine the most frequent ancestry for each sample
        most_freq_ancestry_sample = np.apply_along_axis(
            lambda row: np.bincount(row).argmax(), axis=1, arr=flat_ancestry_pairs
        )
        
        # Obtain unique ancestry values and total counts
        ancestry_values, ancestry_counts = np.unique(lai_T, return_counts=True)
        # Sort ancestry values by total counts in decreasing order
        sorted_ancestry_values = ancestry_values[np.argsort(ancestry_counts)[::-1]]        
        
        # For each ancestry, obtain the samples where that ancestry is predominant
        # and store the indexes sorted in decreasing order based on the ancestry count
        all_sorted_row_idxs = []
        for ancestry in sorted_ancestry_values:
            ancestry_filter = np.where(most_freq_ancestry_sample == ancestry)[0]
            ancestry_counts_sample = np.sum(
                flat_ancestry_pairs[ancestry_filter, :]==ancestry, axis=1
            )
            sorted_row_idxs_1 = np.argsort(ancestry_counts_sample)[::-1]
            sorted_row_idxs = ancestry_filter[sorted_row_idxs_1]
            all_sorted_row_idxs += list(sorted_row_idxs)
        
        # Sort sample IDs based on most frequent ancestry
        if laiobj.samples is not None:
            sample_ids = [laiobj.samples[idx] for idx in all_sorted_row_idxs]
        else:
            sample_ids = None
        
        # Sort `lai_T` based on most frequent ancestry
        num_samples, num_maternal_paternal, num_windows = cast(Tuple[int, int, int], maternal_paternal_pairs.shape)
        lai_T = maternal_paternal_pairs[all_sorted_row_idxs, :].reshape(
            -1, num_windows
        )
    
    # Check if ancestry_map keys are strings of integers (reversed form)
    if all(isinstance(key, str) and key.isdigit() for key in laiobj.ancestry_map.keys()):
        # Reverse the dictionary to match integer-to-ancestry format
        ancestry_map_reverse = {int(key): value for key, value in laiobj.ancestry_map.items()}
    else:
        # The ancestry_map is already in the correct integer-to-ancestry format
        ancestry_map_reverse = laiobj.ancestry_map
    
    # Dictionary with integer-to-color mapping
    colors_map = {key : colors[value] for key, value in ancestry_map_reverse.items()}
    
    # Total number of ancestries
    n_ancestries = len(ancestry_map_reverse.keys())
    
    # Insert whitespace between samples
    lai_T_with_whitespace = np.insert(lai_T, np.arange(2, n_samples*2, 2), n_ancestries, axis=0)
    
    # Repeat rows to increase height of samples in plot
    lai_T_repeat = np.repeat(lai_T_with_whitespace, scale, axis=0)
    
    colors_map[n_ancestries] = 'white'
    
    # Configure custom map from matrix values to colors
    cmap = _custom_cmap(colors_map)
    
    # Plot LAI output
    if figsize is None:
        plt.figure(figsize=(25, 25))
    else:
        plt.figure(figsize=figsize)
    plt.imshow(lai_T_repeat, cmap=cmap)
    
    # Display sample IDs in y-axis
    yticks_positions = np.arange(scale, lai_T_repeat.shape[0]+1, scale*(2+1))
    plt.xticks(fontsize=fontsize['xticks'])
    plt.yticks(yticks_positions, sample_ids, fontsize=fontsize['yticks'])
    
    ax = plt.gca()
    ax.set_xlabel('Window', fontsize=fontsize['xlabel'], labelpad=8)
    ax.set_ylabel('Sample', fontsize=fontsize['ylabel'])
    ax.spines['top'].set_linewidth(2)
    ax.spines['right'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(2)
    ax.spines['left'].set_linewidth(2)
    ax.tick_params(axis='both', which='major', length=8, width=2)
    ax.tick_params(axis='x', which='major', pad=4)
    ax.tick_params(axis='y', which='major', pad=4)
    
    if legend:
        if sort:
            # Sort legend based on global ancestry frequency in decresaing order 
            sorted_colors = [colors_map[x] for x in sorted_ancestry_values]
            ancestries = [ancestry_map_reverse[x] for x in sorted_ancestry_values]
            legend_patches = [patches.Patch(color=color, label=label) 
                              for color, label in zip(sorted_colors, ancestries)]
        else:
            # Add patches for each color to represent the legend squares
            legend_patches = [patches.Patch(color=color, label=label) 
                              for color, label in zip(colors.values(), colors.keys())]
    
        ax.legend(
            handles=legend_patches, 
            loc='lower center', 
            borderaxespad=-5,
            ncol=n_ancestries,
            fontsize=fontsize['legend']
        )
    
    if title:
        plt.title(title)



import tempfile
import pytest
import numpy as np
from pathlib import Path
import os

from snputils.snp.io.read import BEDReader, PGENReader


@pytest.fixture
def test_bed_files():
    """Create test BED files with non-numeric phenotypes including text values."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        # Create a test FAM file with various non-numeric phenotype values
        fam_lines = [
            "FAM1 IND1 0 0 1 -9",     # Missing (PLINK format)
            "FAM2 IND2 0 0 2 NA",     # Missing (common alternative)
            "FAM3 IND3 0 0 0 0",      # Missing (alternative format)
            "FAM4 IND4 0 0 1 1",      # Control
            "FAM5 IND5 0 0 2 2",      # Case
            "FAM6 IND6 0 0 1 case",   # Text value - categorical
        ]

        # Create minimal BIM file with 2 variants
        bim_lines = [
            "1 SNP1 0 1000 A G",
            "1 SNP2 0 2000 C T",
        ]

        # Set up paths
        base_path = tmp_path / "test"
        bed_base = str(base_path)

        fam_path = base_path.with_suffix(".fam")
        bim_path = base_path.with_suffix(".bim")
        bed_path = base_path.with_suffix(".bed")

        # Write FAM and BIM files
        with open(fam_path, "w") as f:
            f.write("\n".join(fam_lines))

        with open(bim_path, "w") as f:
            f.write("\n".join(bim_lines))

        # Create a minimal BED file (binary format)
        # (actual content doesn't matter since we're only testing phenotype handling)
        with open(bed_path, "wb") as f:
            f.write(b"\x6c\x1b\x01\x00")  # Magic bytes + minimal data

        yield bed_base


@pytest.fixture
def test_pgen_files():
    """Create minimal test PGEN fileset with non-numeric phenotypes."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        pgen_base = str(tmp_path / "test_pgen")

        # 1. Create PSAM file with various phenotype values
        psam_lines = [
            "#FID\tIID\tSEX\tPHENO1",
            "FAM1\tIND1\t1\t-9",      # Missing (PLINK format)
            "FAM2\tIND2\t2\tNA",      # Missing (common alternative)
            "FAM3\tIND3\t0\t0",       # Missing (alternative format)
            "FAM4\tIND4\t1\t1",       # Control
            "FAM5\tIND5\t2\t2",       # Case
            "FAM6\tIND6\t1\tcase",    # Text value - categorical
        ]

        # 2. Create a minimal PVAR file
        pvar_lines = [
            "##fileformat=VCFv4.2",
            "#CHROM\tPOS\tID\tREF\tALT",
            "1\t1000\tSNP1\tG\tA",
            "1\t2000\tSNP2\tT\tC"
        ]

        # 3. Create an empty PGEN file (content doesn't matter since we're not reading genotypes)
        pgen_path = Path(pgen_base + ".pgen")
        pvar_path = Path(pgen_base + ".pvar")
        psam_path = Path(pgen_base + ".psam")

        # Make directory if needed
        os.makedirs(os.path.dirname(pgen_path), exist_ok=True)

        # Write the files
        with open(psam_path, "w") as f:
            f.write("\n".join(psam_lines))

        with open(pvar_path, "w") as f:
            f.write("\n".join(pvar_lines))

        with open(pgen_path, "wb") as f:
            f.write(b"\0")  # Just a placeholder

        yield pgen_base


def test_bed_text_phenotypes(test_bed_files):
    """Test that BEDReader can handle text phenotype values."""
    reader = BEDReader(test_bed_files)

    # Test reading sample IDs
    snpobj = reader.read(fields=["IID"])
    assert snpobj.samples is not None
    assert len(snpobj.samples) == 6
    expected_samples = ["IND1", "IND2", "IND3", "IND4", "IND5", "IND6"]
    assert np.array_equal(snpobj.samples, np.array(expected_samples))


def test_pgen_text_phenotypes(test_pgen_files):
    """Test that PGENReader can handle text phenotype values."""
    reader = PGENReader(test_pgen_files)

    # Test reading sample IDs
    snpobj = reader.read(fields=["IID"])
    assert snpobj.samples is not None
    assert len(snpobj.samples) == 6
    expected_samples = ["IND1", "IND2", "IND3", "IND4", "IND5", "IND6"]
    assert np.array_equal(snpobj.samples, np.array(expected_samples))


def test_bed_filter_text_phenotype_samples(test_bed_files):
    """Test filtering samples with text phenotypes in BED files."""
    reader = BEDReader(test_bed_files)

    # Include the sample with text phenotype ("case")
    sample_subset = ["IND1", "IND5", "IND6"]  # IND6 has text phenotype
    snpobj = reader.read(fields=["IID"], sample_ids=sample_subset)

    # Verify filtered samples
    assert snpobj.samples is not None
    assert len(snpobj.samples) == 3
    assert set(snpobj.samples) == set(sample_subset)


def test_pgen_filter_text_phenotype_samples(test_pgen_files):
    """Test filtering samples with text phenotypes in PGEN files."""
    reader = PGENReader(test_pgen_files)

    # Include the sample with text phenotype ("case")
    sample_subset = ["IND1", "IND5", "IND6"]  # IND6 has text phenotype
    snpobj = reader.read(fields=["IID"], sample_ids=sample_subset)

    # Verify filtered samples
    assert snpobj.samples is not None
    assert len(snpobj.samples) == 3
    assert set(snpobj.samples) == set(sample_subset)



from pathlib import Path
from typing import Union

from snputils.ancestry.genobj.local import LocalAncestryObject


def read_lai(file: Union[str, Path], **kwargs) -> LocalAncestryObject:
    """
    Automatically detect the local ancestry data file format from the file's extension and 
    read it into a `snputils.ancestry.genobj.LocalAncestryObject`.

    **Supported formats:**

    - `.msp`: Text-based MSP format.
    - `.msp.tsv`: Text-based MSP format with TSV extension.
    
    Args:
        file (str or pathlib.Path): 
            Path to the file to be read. It should end with `.msp` or `.msp.tsv`.
        **kwargs: Additional arguments passed to the reader method.
    """
    from snputils.ancestry.io.local.read.auto import LAIReader

    return LAIReader(file).read(**kwargs)


def read_msp(file: Union[str, Path]) -> 'LocalAncestryObject':
    """
    Read data from an `.msp` or `.msp.tsv` file and construct a `snputils.ancestry.genobj.LocalAncestryObject`.

    Args:
        file (str or pathlib.Path): 
            Path to the file to be read. It should end with `.msp` or `.msp.tsv`.

    Returns:
        **LocalAncestryObject:**
            A LocalAncestryObject instance.
    """
    from snputils.ancestry.io.local.read.msp import MSPReader

    return MSPReader(file).read()



import inspect
import os
import pathlib
import subprocess
import urllib.request
import zipfile
import platform

import pytest

import snputils
from snputils.snp.io.read import BEDReader, PGENReader, VCFReader


@pytest.fixture(scope="module")
def data_path():
    module_path = pathlib.Path(inspect.getfile(snputils)).parent.parent
    data_path = module_path / "data"
    os.makedirs(data_path, exist_ok=True)

    system = platform.system()
    is_arm = platform.machine().lower().startswith(('arm', 'aarch'))

    plink_urls = {
        ("Darwin", True): "plink2_mac_arm64_20241114.zip",  # macOS ARM
        ("Linux", False): "plink2_linux_x86_64_20241114.zip",  # Linux x86_64
    }

    try:
        plink_filename = plink_urls[(system, is_arm)]
    except KeyError:
        raise RuntimeError(f"Unsupported platform: {system} {platform.machine()}")

    files_urls = {
        plink_filename: f"https://s3.amazonaws.com/plink2-assets/alpha6/{plink_filename}",
        "vcf/ALL.chr22.shapeit2_integrated_v1a.GRCh38.20181129.phased.vcf.gz":
        "https://ftp.1000genomes.ebi.ac.uk/vol1/ftp/data_collections/1000_genomes_project/release/20181203_biallelic_SNV/ALL.chr22.shapeit2_integrated_v1a.GRCh38.20181129.phased.vcf.gz",
    }

    # Check and download each file if it does not exist
    for file_name, url in files_urls.items():
        file_path = data_path / file_name
        dir_path = file_path.parent if file_path.is_file() else file_path
        if not dir_path.exists():
            os.makedirs(file_path.parent, exist_ok=True)
        if not file_path.exists():
            print(f"Downloading {file_name} to {data_path}. This may take a while...")
            urllib.request.urlretrieve(url, file_path)
            if file_path.suffix == ".zip":
                with zipfile.ZipFile(file_path, "r") as zip_ref:
                    zip_ref.extractall(data_path)
                subprocess.run(
                    ["chmod", "+x", data_path / "plink2"], cwd=str(data_path)
                )

        if file_path.suffixes[-2:] == ['.vcf', '.gz']:
            # Subset sample files
            subset_file = data_path / "subset.txt"
            if not subset_file.exists():
                print("Generating subset file...")
                four_sample_ids = ["HG00096", "HG00097", "HG00099", "HG00100"]
                with open(subset_file, "w") as file:
                    file.write("\n".join(four_sample_ids))

            plink_vcf_out = "vcf/subset"
            subset_vcf_file = data_path / (plink_vcf_out + ".vcf")
            if not subset_vcf_file.exists():
                print("Generating subset VCF...")
                subprocess.run(
                    [
                        "./plink2",
                        "--vcf",
                        data_path / file_name,
                        "--keep",
                        "subset.txt",
                        "--recode",
                        "vcf",
                        "--set-missing-var-ids",
                        "@:#",
                        "--out",
                        plink_vcf_out,
                    ],
                    cwd=str(data_path),
                )

    # Generate bed and pgen formats
    for fmt in ["bed", "pgen", "pgen_zst"]:
        fmt_path = data_path / fmt
        os.makedirs(fmt_path, exist_ok=True)
        fmt_file = fmt_path / "subset"
        if not fmt_file.exists():
            print(f"Generating {fmt} format...")
            make_fmt = "--make-pgen vzs" if fmt == "pgen_zst" else f"--make-{fmt}"
            subprocess.run(
                [
                    "./plink2",
                    "--vcf",
                    subset_vcf_file,
                    *make_fmt.split(),
                    "--out",
                    fmt_file,
                ],
                cwd=str(data_path),
            )

    return str(data_path)


@pytest.fixture(scope="module")
def snpobj_vcf(data_path):
    return VCFReader(data_path + "/vcf/subset.vcf").read(sum_strands=False)


@pytest.fixture(scope="module")
def snpobj_bed(data_path):
    return BEDReader(data_path + "/bed/subset").read(sum_strands=False)


@pytest.fixture(scope="module")
def snpobj_pgen(data_path):
    return PGENReader(data_path + "/pgen/subset").read(sum_strands=False)



import copy
import warnings
import numpy as np
import pandas as pd
from typing import Union, Sequence, Optional


class MultiPhenotypeObject():
    """
    A class for multi-phenotype data.

    This class serves as a container for phenotype data, allowing for
    operations such as filtering samples and accessing phenotype information.
    It uses a DataFrame to store the data, with the first column reserved for the sample identifers.
    """
    def __init__(
        self,
        phen_df: pd.DataFrame
    ) -> None:
        """
        Args:
            phen_df (pd.DataFrame): 
                A Pandas DataFrame containing phenotype data, with the first column 
                representing sample identifiers.
        """
        self.__phen_df = phen_df

    def __getitem__(self, key):
        """
        To access an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            return getattr(self, key)
        except:
            raise KeyError(f'Invalid key: {key}')

    def __setitem__(self, key, value):
        """
        To set an attribute of the class using the square bracket notation,
        similar to a dictionary.
        """
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(f'Invalid key: {key}')

    @property
    def phen_df(self) -> pd.DataFrame:
        """
        Retrieve `phen_df`.

        Returns:
            pd.DataFrame: 
                A Pandas DataFrame containing phenotype data, with the first column 
                representing sample identifiers.
        """
        return self.__phen_df
    
    @phen_df.setter
    def phen_df(self, x: pd.DataFrame):
        """
        Update `phen_df`.
        """
        self.__phen_df = x
    
    @property
    def n_samples(self) -> int:
        """
        Retrieve `n_samples`.

        Returns:
            int: The total number of samples.
        """
        return len(self.phen_df)

    def copy(self):
        """
        Create and return a copy of the current `MultiPhenotypeObject` instance.

        Returns:
            MultiPhenotypeObject: A new instance of the current object.
        """
        return copy.copy(self)
    
    def filter_samples(
            self, 
            samples: Optional[Union[str, Sequence[str], np.ndarray]] = None, 
            indexes: Optional[Union[int, Sequence[int], np.ndarray]] = None, 
            include: bool = True, 
            reorder: bool = False, 
            inplace: bool = False
        ) -> Optional['MultiPhenotypeObject']:
        """
        Filter samples in the `MultiPhenotypeObject` based on sample names or indexes.

        This method allows you to include or exclude specific samples by their names,
        indexes, or both. When both samples and indexes are provided, the union of
        the specified samples is used. Negative indexes are supported and follow NumPy's indexing 
        conventions. Set `reorder=True` to match the ordering of the provided `samples` and/or
        `indexes` lists when including.

        Args:
            samples (str or array_like of str, optional): 
                 Names of the samples to include or exclude. Can be a single sample name or a
                 sequence of sample names. Default is None.
            indexes (int or array_like of int, optional):
                Indexes of the samples to include or exclude. Can be a single index or a sequence
                of indexes. Negative indexes are supported. Default is None.
            include (bool, default=True): 
                If True, includes only the specified samples. If False, excludes the specified
                samples. Default is True.
            inplace (bool, default=False): 
                If True, modifies the object in place. If False, returns a new
                `MultiPhenotypeObject` with the samples filtered. Default is False.

        Returns:
            Optional[MultiPhenotypeObject]: Returns a new MultiPhenotypeObject with the specified samples 
            filtered if `inplace=False`. If `inplace=True`, modifies the object in place and returns None.
        """
        # Ensure at least one of samples or indexes is provided
        if samples is None and indexes is None:
            raise ValueError("At least one of 'samples' or 'indexes' must be provided.")

        n_samples = self.n_samples

        # Create mask based on sample names
        if samples is not None:
            samples = np.asarray(samples).ravel()
            # Extract sample names from the DataFrame
            sample_names = self.__phen_df.iloc[:, 0].values
            # Create mask for samples belonging to specified names
            mask_samples = np.isin(sample_names, samples)
        else:
            mask_samples = np.zeros(n_samples, dtype=bool)

        # Create mask based on sample indexes
        if indexes is not None:
            indexes = np.asarray(indexes).ravel()
            # Adjust negative indexes
            indexes = np.mod(indexes, n_samples)
            if np.any((indexes < 0) | (indexes >= n_samples)):
                raise IndexError("One or more sample indexes are out of bounds.")
            # Create mask for samples at specified indexes
            mask_indexes = np.zeros(n_samples, dtype=bool)
            mask_indexes[indexes] = True
        else:
            mask_indexes = np.zeros(n_samples, dtype=bool)

        # Combine masks using logical OR (union of samples)
        mask_combined = mask_samples | mask_indexes

        if not include:
            # Invert mask if excluding samples
            mask_combined = ~mask_combined

        # If requested, compute an ordering of selected rows that follows the provided lists
        ordered_indices = None
        if include and reorder:
            sel_indices = np.where(mask_combined)[0]
            sample_names = self.__phen_df.iloc[:, 0].values
            ordered_list = []
            added = np.zeros(n_samples, dtype=bool)

            # Respect the order provided in `samples` (supports duplicate sample names)
            if samples is not None:
                for s in samples:
                    matches = np.where(sample_names == s)[0]
                    for idx in matches:
                        if mask_combined[idx] and not added[idx]:
                            ordered_list.append(int(idx))
                            added[idx] = True

            # Then respect the order in `indexes`
            if indexes is not None:
                adj_idx = np.mod(np.atleast_1d(indexes), n_samples)
                for idx in adj_idx:
                    if mask_combined[idx] and not added[idx]:
                        ordered_list.append(int(idx))
                        added[idx] = True

            # Finally, append any remaining selected rows in their original order
            for idx in sel_indices:
                if not added[idx]:
                    ordered_list.append(int(idx))

            ordered_indices = np.asarray(ordered_list, dtype=int)

        # Filter the phenotype DataFrame
        if inplace:
            if ordered_indices is not None:
                self['phen_df'] = self['phen_df'].iloc[ordered_indices].reset_index(drop=True)
            else:
                self['phen_df'] = self['phen_df'][mask_combined].reset_index(drop=True)
            return None
        else:
            phen_obj = self.copy()
            if ordered_indices is not None:
                phen_obj['phen_df'] = phen_obj['phen_df'].iloc[ordered_indices].reset_index(drop=True)
            else:
                phen_obj['phen_df'] = phen_obj['phen_df'][mask_combined].reset_index(drop=True)
            return phen_obj



import logging
from typing import List, Optional
import os
import csv

import numpy as np
import polars as pl
import pgenlib as pg

from snputils.snp.genobj.snpobj import SNPObject
from snputils.snp.io.read.base import SNPBaseReader

log = logging.getLogger(__name__)


@SNPBaseReader.register
class PGENReader(SNPBaseReader):
    def read(
        self,
        fields: Optional[List[str]] = None,
        exclude_fields: Optional[List[str]] = None,
        sample_ids: Optional[np.ndarray] = None,
        sample_idxs: Optional[np.ndarray] = None,
        variant_ids: Optional[np.ndarray] = None,
        variant_idxs: Optional[np.ndarray] = None,
        sum_strands: bool = False,
        separator: str = None,
    ) -> SNPObject:
        """
        Read a pgen fileset (pgen, psam, pvar) into a SNPObject.

        Args:
            fields (str, None, or list of str, optional): Fields to extract data for that should be included in the returned SNPObject.
                Available fields are 'GT', 'IID', 'REF', 'ALT', '#CHROM', 'ID', 'POS', 'FILTER', 'QUAL'.
                To extract all fields, set fields to None. Defaults to None.
            exclude_fields (str, None, or list of str, optional): Fields to exclude from the returned SNPObject.
                Available fields are 'GT', 'IID', 'REF', 'ALT', '#CHROM', 'ID', 'POS', 'FILTER', 'QUAL'.
                To exclude no fields, set exclude_fields to None. Defaults to None.
            sample_ids: List of sample IDs to read. If None and sample_idxs is None, all samples are read.
            sample_idxs: List of sample indices to read. If None and sample_ids is None, all samples are read.
            variant_ids: List of variant IDs to read. If None and variant_idxs is None, all variants are read.
            variant_idxs: List of variant indices to read. If None and variant_ids is None, all variants are read.
            sum_strands: If True, maternal and paternal strands are combined into a single `int8` array with values `{0, 1, 2`}. 
                If False, strands are stored separately as an `int8` array with values `{0, 1}` for each strand. 
                Note: With the pgenlib backend, `False` uses `~8×` more RAM, though `calldata_gt` is only `2×` larger.
            separator: Separator used in the pvar file. If None, the separator is automatically detected.
                If the automatic detection fails, please specify the separator manually.

        Returns:
            **SNPObject**: 
                A SNPObject instance.
        """
        assert (
            sample_idxs is None or sample_ids is None
        ), "Only one of sample_idxs and sample_ids can be specified"
        assert (
            variant_idxs is None or variant_ids is None
        ), "Only one of variant_idxs and variant_ids can be specified"

        if isinstance(fields, str):
            fields = [fields]
        if isinstance(exclude_fields, str):
            exclude_fields = [exclude_fields]

        fields = fields or ["GT", "IID", "REF", "ALT", "#CHROM", "ID", "POS", "FILTER", "QUAL"]
        exclude_fields = exclude_fields or []
        fields = [field for field in fields if field not in exclude_fields]
        only_read_pgen = fields == ["GT"] and variant_idxs is None and sample_idxs is None

        filename_noext = str(self.filename)
        for ext in [".pgen", ".pvar", ".pvar.zst", ".psam"]:
            if filename_noext.endswith(ext):
                filename_noext = filename_noext[:-len(ext)]
                break

        if only_read_pgen:
            file_num_samples = None  # Not needed for pgen
            file_num_variants = None  # Not needed
        else:
            pvar_extensions = [".pvar", ".pvar.zst"]
            pvar_filename = None
            for ext in pvar_extensions:
                possible_pvar = filename_noext + ext
                if os.path.exists(possible_pvar):
                    pvar_filename = possible_pvar
                    break
            if pvar_filename is None:
                raise FileNotFoundError(f"No .pvar or .pvar.zst file found for {filename_noext}")

            log.info(f"Reading {pvar_filename}")

            def open_textfile(filename):
                if filename.endswith('.zst'):
                    import zstandard as zstd
                    return zstd.open(filename, 'rt')
                else:
                    return open(filename, 'rt')

            pvar_has_header = True
            pvar_header_line_num = 0
            with open_textfile(pvar_filename) as file:
                for line_num, line in enumerate(file):
                    if line.startswith("##"):  # Metadata
                        continue
                    else:
                        if separator is None:
                            separator = csv.Sniffer().sniff(file.readline()).delimiter
                        if line.startswith("#CHROM"):  # Header
                            pvar_header_line_num = line_num
                            header = line.strip().split()
                            break
                        elif not line.startswith("#"):  # If no header, look at line 1
                            pvar_has_header = False
                            cols_in_pvar = len(line.strip().split(separator))
                            if cols_in_pvar == 5:
                                header = ["#CHROM", "ID", "POS", "ALT", "REF"]
                            elif cols_in_pvar == 6:
                                header = ["#CHROM", "ID", "CM", "POS", "ALT", "REF"]
                            else:
                                raise ValueError(
                                    f"{pvar_filename} is not a valid pvar file."
                                )
                            break

            pvar_reading_args = {
                'separator': separator,
                'skip_rows': pvar_header_line_num,
                'has_header': pvar_has_header,
                'new_columns': None if pvar_has_header else header,
                'schema_overrides': {
                    "#CHROM": pl.String,
                    "POS": pl.UInt32,
                    "ID": pl.String,
                    "REF": pl.String,
                    "ALT": pl.String,
                },
                'null_values': ["NA"],
            }
            if pvar_filename.endswith('.zst'):
                pvar = pl.read_csv(pvar_filename, **pvar_reading_args).lazy()
            else:
                pvar = pl.scan_csv(pvar_filename, **pvar_reading_args)

            # We need to track row positions of variants but doing this with lazy operations would
            # defeat predicate pushdown optimization and force loading the entire DataFrame.
            # Solution: We only materialize the ID column with indices, keeping memory usage minimal
            # while maintaining the ability to map between variant IDs and their file positions.
            variant_id_idxs = pvar.select("ID").with_row_index().collect()

            file_num_variants = variant_id_idxs.height

            if variant_ids is not None:
                num_variants = np.size(variant_ids)
                pvar = pvar.filter(pl.col("ID").is_in(variant_ids)).collect()
                variant_idxs = variant_id_idxs.filter(pl.col("ID").is_in(variant_ids)).select("index").to_series().to_numpy()
            elif variant_idxs is not None:
                num_variants = np.size(variant_idxs)
                variant_idxs = np.array(variant_idxs, dtype=np.uint32)
                variant_ids = variant_id_idxs.filter(pl.col("index").is_in(variant_idxs)).select("ID")
                pvar = pvar.filter(pl.col("ID").is_in(variant_ids)).collect()
            else:
                num_variants = file_num_variants
                variant_idxs = np.arange(num_variants, dtype=np.uint32)
                pvar = pvar.collect()

            log.info(f"Reading {filename_noext}.psam")

            with open(filename_noext + ".psam") as file:
                first_line = file.readline().strip()
                psam_has_header = first_line.startswith(("#FID", "FID", "#IID", "IID"))

            psam = pl.read_csv(
                filename_noext + ".psam",
                separator=separator,
                has_header=psam_has_header,
                new_columns=None if psam_has_header else ["FID", "IID", "PAT", "MAT", "SEX", "PHENO1"],
                null_values=["NA"],
            ).with_row_index()
            if "#IID" in psam.columns:
                psam = psam.rename({"#IID": "IID"})
            if "#FID" in psam.columns:
                psam = psam.rename({"#FID": "FID"})

            file_num_samples = psam.height

            if sample_ids is not None:
                num_samples = np.size(sample_ids)
                psam = psam.filter(pl.col("IID").is_in(sample_ids))
                sample_idxs = psam.select("index").to_series().to_numpy()
            elif sample_idxs is not None:
                num_samples = np.size(sample_idxs)
                sample_idxs = np.array(sample_idxs, dtype=np.uint32)
                psam = psam.filter(pl.col("index").is_in(sample_idxs))
            else:
                num_samples = file_num_samples

        if "GT" in fields:
            log.info(f"Reading {filename_noext}.pgen")
            pgen_reader = pg.PgenReader(
                str.encode(filename_noext + ".pgen"),
                raw_sample_ct=file_num_samples,
                variant_ct=file_num_variants,
                sample_subset=sample_idxs,
            )

            if only_read_pgen:
                num_samples = pgen_reader.get_raw_sample_ct()
                num_variants = pgen_reader.get_variant_ct()
                variant_idxs = np.arange(num_variants, dtype=np.uint32)

            # required arrays: variant_idxs + sample_idxs + genotypes
            if not sum_strands:
                required_ram = (num_samples + num_variants + num_variants * 2 * num_samples) * 4
            else:
                required_ram = (num_samples + num_variants) * 4 + num_variants * num_samples
            log.info(f">{required_ram / 1024**3:.2f} GiB of RAM are required to process {num_samples} samples with {num_variants} variants each")

            if not sum_strands:
                genotypes = np.empty((num_variants, 2 * num_samples), dtype=np.int32)  # cannot use int8 because of pgenlib
                pgen_reader.read_alleles_list(variant_idxs, genotypes)
                genotypes = genotypes.astype(np.int8).reshape((num_variants, num_samples, 2))
            else:
                genotypes = np.empty((num_variants, num_samples), dtype=np.int8)
                pgen_reader.read_list(variant_idxs, genotypes)
            pgen_reader.close()
        else:
            genotypes = None

        log.info("Constructing SNPObject")

        snpobj = SNPObject(
            calldata_gt=genotypes if "GT" in fields else None,
            samples=psam.get_column("IID").to_numpy() if "IID" in fields and "IID" in psam.columns else None,
            **{f'variants_{k.lower()}': pvar.get_column(v).to_numpy() if v in fields and v in pvar.columns else None
               for k, v in {'ref': 'REF', 'alt': 'ALT', 'chrom': '#CHROM', 'id': 'ID', 'pos': 'POS', 'filter_pass': 'FILTER', 'qual': 'QUAL'}.items()}
        )

        log.info("Finished constructing SNPObject")
        return snpobj
